import os
import sys
import glob
import csv
from math import exp, log, sin, cos, tan, log
import numpy as np
from numpy import sin, cos, tan, pi
from scipy.interpolate import interp1d
from pprint import pprint


import tklib.tkre
from tklib.tkobject import tkObject
from tklib.tkprogvars import ProgramDir, DBDir, AtomDBDir
from tklib.tkutils import IsDir, IsFile, pint, pfloat
from tklib.tkfile import tkFile
from tklib.tkcrystal.tkatomtypeobject import tkAtomTypeObject
from tklib.tksci.tksci import a0, torad, todeg, Round


ASFDCPath      = os.path.join(DBDir, 'asfdc')
ScantteringDir = os.path.join(DBDir, 'atom_scattering', 'data')


#=====================================
# Space group class
#=====================================

class tkAtomType(tkAtomTypeObject):
    def __init__(self, atomtype = None, charge = 0.0, **args):
        super(tkAtomType, self).__init__(atomtype, charge,**args)
        self.__ASFDCParameters = None
        self.__ASFDCAtomName   = None
        self.__MACE  = None
        self.__MACwl = None
        self.__MACf1 = None
        self.__MACf2 = None
        self.update(**args)

    def __del__(self):
        super(tkAtomType, self).__del__()

    def __str__(self):
        return self.ClassPath()

    def ASFDCParameters(self):
        return self.__ASFDCParameters

    def ASFDCAtomName(self):
        return self.__ASFDCAtomName

    def ReadASFParameters(self, AtomName = None, asfdcfile = ASFDCPath, XraySource = None, StopByError = 1):
        SetAttribute = 0
        if AtomName is None:
            AtomName = self.AtomType()
            SetAttribute = 1

#        atomname, nametype, sitetype, charge = self.SplitAtomName(AtomName)
#        atomname = atomname.upper()
        atomname = AtomName.upper()

        fp = tkFile(asfdcfile)
        if fp is None:
            if StopByError:
                print("")
                print("Error in tkcrystal.tkatomtype.ReadASFParameters:")
                print("  Can not read [{}]".format(asfdcfile))
                print("")
                exit()
            return None

        ReadAtomName = 0
        while 1:
            line = fp.ReadLine()
            if not line:
                break
        
            aa = line.split()
            name = aa[0]
            if name.upper() == atomname:
                ReadAtomName = 1
                break
#        print("Atom: [{}] {}".format(AtomName, ReadAtomName))
        if not ReadAtomName:
            if StopByError:
                print("")
                print("Error in tkcrystal.tkatomtype.ReadASFParameters:")
                print("  Can not find [{}] in [{}]".format(atomname, asfdcfile))
                print("")
                exit()
            return None

# f(s = sin Q / l) = $C + sum_{i=1, 5}[Ai * exp(-Bi * s^2)] 
#  if @a1 ==  9: line1 = ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $C, $b)    $A5 = 0.0, $B5 = 1.0
#  if @a1 == 11: line2 = ($A1, $A2, $A3, $A4, $A5, $C, $B1, $B2, $B3, $B4, $B5, $b)
        line1 = fp.ReadLine()
        line2 = fp.ReadLine()
        a1    = line1.split()
        a2    = line2.split()
        for i in range(len(a1)):
            a1[i] = float(a1[i])
        for i in range(len(a2)):
            a2[i] = float(a2[i])

        if len(a2) == 11:
            a1 = [a2[0], a2[6], a2[1], a2[7], a2[2], a2[8], a2[3], a2[9], a2[4], a2[10], a2[5]]
# read skip one more line
            line2 = fp.ReadLine()
        else:
            a1 = [a1[0], a1[1], a1[2], a1[3], a1[4], a1[5], a1[6], a1[7], 0.0, 1.0, a1[8]]

        a3 = []
        if '0/' not in line2:
            a3 = line2.split()

        inf = self.GetAtomInformation(atomname.capitalize())

        fp.Close()

        if XraySource is not None and type(XraySource) is float:
            f1, f2 = self.AnomalousScatteringFactor(XraySource)
            a1 = a1 + [f1, f2]
        elif XraySource is not None:
            A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C = a1
            Crdf1, Crdf2, Fedf1, Fedf2, Codf1, Codf2, Cudf1, Cudf2, Modf1, Modf2, Agdf1, Agdf2, Kbdf1, Kbdf2 = a3

            if XraySource == 'Cr' and Crdf2 is not None:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C, Crdf1, Crdf2]
            elif XraySource == 'Co' and Codf2 is not None:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C, Codf1, Codf2]
            elif XraySource == 'Cu' and Cudf2 is not None:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C, Cudf1, Cudf2]
            elif XraySource == 'Mo' and Modf2 is not None:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C, Modf1, Modf2]
            elif XraySource == 'Ag' and Agdf2 is not None:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C, Agdf1, Agdf2]
            elif XraySource == 'Kb' and AKbgdf2 is not None:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C, Kbdf1, Kbdf2]
            else:
                a1 = [A1, B1, A2, B2, A3, B3, A4, B4, A5, B5, C]

        if SetAttribute:
            self.__ASFDCParameters = a1
            self.__ASFDCAtomName    = atomname

        return a1

# $s = sin Q / lambda (in nm)
    def asf(self, s, StopByError = 1):
        s2 = 0.01 * s * s                   # convert to anstrom^-2

        a1  = self.ASFDCParameters()
        if a1 is None:
            a1 = self.ReadASFParameters()
        n   = 11
        asf = a1[n-1]
        for i in range(0, n-1, 2):
            asf += a1[i] * exp(-a1[i+1] * s2)

        try:
            return asf + a1[11] + 1.0j * a1[12]
        except:
            return asf

    # data taken from http://skuld.bmsc.washington.edu/scatter/
    # E in eV, wavelength in nm
    def ReadAnomalousScatteringFactor(self, AtomName = None):
        SetAttribute = 0
        if AtomName is None:
            AtomName = self.AtomType()
            SetAttribute = 1

        atomname, nametype, sitetype, charge = self.SplitAtomName(AtomName)
        atomname = atomname.capitalize()

        dbpath = os.path.join(ScantteringDir, atomname + ".dat")

        xE, yf1, yf2 = np.genfromtxt(dbpath, skip_header = 0, usecols = (0, 1, 2), unpack = True)
        xwl = []
        for i in range(len(xE)):
            xwl.append(1239.758636 / xE[i])

        if SetAttribute:
            self.__MACE  = xE
            self.__MACwl = xwl
            self.__MACf1 = yf1
            self.__MACf2 = yf2

        return xE, xwl, yf1, yf2, dbpath

    def AnomalousScatteringFactor(self, wl):
        if self.__MACf1 is None:
            self.ReadAnomalousScatteringFactor()

        func1 = interp1d(self.__MACwl, self.__MACf1, kind = 'linear')
        f1 = func1(wl) + 0.0
        func2 = interp1d(self.__MACwl, self.__MACf2, kind = 'linear')
        f2 = func2(wl) + 0.0
        return f1, f2


"""
#my $DBDir = Deps::MakePath($WebRootDir, "Research", 0);
#   $DBDir = Deps::MakePath($DBDir, "Databases", 0);

my $NonrelInputPath           = Deps::MakePath($DBDir, "nonrel",                      0);
my $IonRadius1DBPath          = Deps::MakePath($DBDir, "IonRadius1.html",             0);
my $IonRadius2DBPath          = Deps::MakePath($DBDir, "IonRadius2.html",             0);
my $PeriodicTable1DBPath      = Deps::MakePath($DBDir, "PeriodicTable1.html",         0);
my $PeriodicTable2DBDir       = Deps::MakePath($DBDir, "PeriodicTable2",              0);
my $PeriodicTable3DB          = Deps::MakePath($DBDir, "PERIODIC.CSV",                0);
my $ElectronicConfigurationDB = Deps::MakePath($DBDir, "ElectronicConfiguration.csv", 0);

my $MXDDir     = ProgVars::MXDDir();
my $MXDExecDir = ProgVars::MXDExecDir();
my $SX1DBPath  = ProgVars::SX1DBPath();

sub GetAtomDBHash {
	my ($element, $charge, $nCoordination) = @_;
	$charge = 0 if(!defined $charge);

	if($element =~ /^\d+$/) {
		$element = AtomType::GetAtomName($element);
	}

	my ($mass, $charge0, $ai, $bi, $ci, $rad) = AtomType::GetBusingParameter($element);
	my ($AtmNum, $JName, $EName, $NameOrigin) = AtomType::GetIonInfFromPeriodicTable1($element);
	my ($AtomicNumber, $AtomName, $Charge)    = AtomType::GetAtomInformation($element);
	my $CationRadius = AtomType::GetCationRadius($element, $charge, $nCoordination);
	my @IonRadius    = AtomType::GetIonRadius2($element, $charge, $nCoordination);
	my ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity) = AtomType::GetIonInfFromPeriodicTable2($AtmNum);
	my $NameOrigin2;
	($AtmNum, $JName, $EName, $NameOrigin2) = AtomType::GetIonInfFromPeriodicTable1($element);
	my ($pOrbitalName1, $pnElectron1)         = AtomType::GetElectronicConfiguraton($AtmNum);
	my ($pOrbitalName2, $pnElectron2)         = AtomType::SpeculateElectronicConfiguration($AtmNum);
	my $ElectronConfiguration1 = '';
	my $ElectronConfiguration2 = '';
	for(my $i = 0 ; $i < @$pOrbitalName1 ; $i++) {
		next if($pnElectron1->[$i] == 0);
		$ElectronConfiguration1 .= "$pOrbitalName1->[$i]($pnElectron1->[$i])";
	}
	for(my $i = 0 ; $i < @$pOrbitalName2 ; $i++) {
		next if($pnElectron2->[$i] == 0);
		$ElectronConfiguration2 .= "$pOrbitalName2->[$i]($pnElectron2->[$i])";
	}
#print "  $ElectronConfiguration1\n";
#print "  $ElectronConfiguration2\n";

	$AtomicMass = $mass        if(!defined $AtomicMass);
	$JName      = $AtomJName   if(!defined $JName);
	$EName      = $AtomEName   if(!defined $EName);
	$NameOrigin = $NameOrigin2 if(!defined $NameOrigin);
	my %db = (
		name                   => $element,
		Z                      => $AtmNum,
		mass                   => $AtomicMass,
		charge                 => $charge,
		nCoordination          => $nCoordination,
		AtomicRdius            => $AtomicRadius,
		CationRadius           => $CationRadius,
		pIonRadius             => \@IonRadius,
		IonizationPotential    => $IonizationPotential,
		ElectronAffinity       => $ElectronAffinity,
		MeltingPoint           => $MeltingPoint,
		BoilingPoint           => $BoilingPoint,
		Density                => $Density,
		HeatCapacity           => $HeatCapacity,
		ElectronConfiguration1 => $ElectronConfiguration1,
		ElectronConfiguration2 => $ElectronConfiguration2,
		ai                     => $ai,
		bi                     => $bi,
		ci                     => $ci,
		rad                    => $rad,
		JName                  => $JName,
		EName                  => $EName,
		NameOrigin             => $NameOrigin,
		FoundYear              => $FoundYear,
		
		);

	return \%db;
}

sub GetShellName
{
	my ($n) = @_;
	my @names = qw(a K L M N O P);
	return $names[$n];
}

sub GetOrbName
{
	my ($n, $l, $m) = @_;
	if(!defined $m) {
		my @names = ('s', 'p', 'd', 'f', 'g', 'h');
		return "$n$names[$l]";
	}

	my @names1 = ('s'); 										# for l=0
	my @names2 = ('py', 'pz', 'px');							# for l=1, m=-1, 0, 1
	my @names3 = ('dxy', 'dyz', 'd3z2-r2', 'dzx', 'dx2-y2');	# for l=2, m=-2,-1,0,1,2
	my @names4 = ('fy(3x2-y2)', 'fxyz', 'fy(5z2-r2)', 'fz(5z2-3r2)',
		      'fx(5z2-r2)', 'fz(x2-y2)', 'fx(x2-3y2)');		# for l=3, m=-3,-2,-1,0,1,2,3
	my @names = (\@names1, \@names2, \@names3, \@names4);

	my $pNames = $names[$l];
	my $name = $pNames->[$m+$l];
	return "$n$name";
}


sub GetEnglishName
{
	my ($name) = (@_);
	my ($AtmNum, $JName, $EName, $NameOrigin)
		= GetIonInfFromPeriodicTable1($name);
	return $EName;
}

sub GetBusingParameter
{
	my ($atomname, $Debug) = @_;

	$atomname =~ s/\{.*?\}//g;
	$atomname =~ s/\d//g;
#print "an: $atomname\n";
#print "DB: $SX1DBPath\n";
	unless(open(IN, "<$SX1DBPath")) {
		print("Error in MXD.pm::GetBusingParameter: "
			."Can not read DB [$SX1DBPath].<br>\n") if($Debug);
		return ('');
	}
	<IN>;
	while(<IN>) {
		my ($name, $mass, $charge, $ai, $bi, $ci, $rad) = split(/\s+/, $_);
#print "uc: $name eq $atomname\n";
		if(uc $name eq uc $atomname) {
			close(IN);
			return ($mass, $charge, $ai, $bi, $ci, $rad);
		}
	}
	
	close(IN);
	return ('');
}

sub GetAtomicNumber 
{
	my ($name, $Debug) = (@_);

	$name =~ s/\{.*?\}//g;
	$name =~ s/\d//g;
	$name = uc $name;
	$name =~ s/^(\D+?)\d.*$/$1/;
	my $an = -1;
	unless(open(IN, "<$NonrelInputPath")) {
		print "Error: $!: $NonrelInputPath${LF}";# if($Debug);
		return -1;
	}

	while(<IN>) {
# H  HFS ATOM CALC.
#    1  300    1
#   1.00000   0.97804  20.00000  32.00000   0.00000
#  1.0  0.0  0.0   -0.30       1.0
  		if($_ =~ /^\s*$name\s/) { # H  HFS ATOM CALC.
			my $line = <IN>;  #    1  300    1
			$line = <IN>;     #   1.00000   0.97804  20.00000
			($an) = ($line =~ /^\s*(.*?)\s/);
			last;
		}
	}
	close(IN);
	return $an;
}

sub ChargeToVal
{
	my ($charge) = @_;

	return  0 if(!defined $charge);
	return  1 if($charge eq '+');
	return -1 if($charge eq '-');
	if($charge =~ /(\d+)\+/ or $charge =~ /\+(\d+)/) {
		return + abs($1);
	}
	if($charge =~ /(\d+)\-/ or $charge =~ /\-(\d+)/) {
		return - abs($1);
	}
	return $charge;
}

sub ChargeToStr
{
	my ($charge) = @_;

	return ''   if(!defined $charge);
	return '1+' if($charge eq '+');
	return '1-' if($charge eq '-');
	if($charge =~ /(\d+)\+/ or $charge =~ /\+(\d+)/) {
		$charge = + abs($1);
	}
	if($charge =~ /(\d+)\-/ or $charge =~ /\-(\d+)/) {
		$charge = - abs($1);
	}
	$charge = '' if($charge == 0.0);
	return $charge;
}

sub GetCationRadius
{
	my ($name, $charge, $nCoordination, $Debug) = (@_);
	$name   = uc $name;

	$charge = &ChargeToVal($charge);

	my $r = -1.0;
	unless(open(IN, "<$IonRadius1DBPath")) {
		print "Error: $!: $IonRadius1DBPath${LF}";# if($Debug);
		return -1;
	}
	while(<IN>) {
		my $line = $_;
		next if($line !~ m|<TR>|i);
		my $atmline    = <IN>;
		my $atmnumline = <IN>;
		my $radiusline = <IN>;
		$atmline =~ m|<TD.*?>(.*?)<SUP>(.*?)<\/SUP><\/TD>|i;
		my $atname   = (defined $1)? uc $1 : '';
		my $atcharge = 0;
		if(defined $2) {
			$atcharge = $2;
		}
#		$atcharge = 0.0 + $atcharge if($atcharge =~ /\+/);
		$atcharge =~ s/\+//;
		if($atcharge =~ /\-/) {
			$atcharge =~ s/-//;
			$atcharge = 0.0 - $atcharge;
		}
		$atmnumline =~ m|<TD.*?>(.*?)<\/TD>|i;
		my $atomicnumber   = $1;
		$radiusline =~ m|<TD.*?>(.*?)<\/TD>|i;
		my $radius   = $1;

		if($name eq $atname and $charge eq $atcharge) {
			close(IN);
			return $radius;
		}
		$r = $radius if($name eq $atname);
	}
	close(IN);
	return $r;
}

sub GetIonRadius2
{
	my ($name, $charge, $nCoord, $Debug) = (@_);

	$name = uc $name;
	my @radius;
	
	unless(open(IN, "<$IonRadius2DBPath")) {
		print "Error: $!: $IonRadius2DBPath${LF}";# if($Debug);
		return -1;
	}
	while(<IN>) {
		my $line = $_;
		$line =~ m|\<P\>\<B\>(.*?)\</B\>\</P\>|i;
		my $AtomName = $1;
		next unless(defined $AtomName and uc $AtomName eq $name);

		my $rline = '';
		while(<IN>) {
			my $line2 = $_;
			last if($line2 =~ m|\<TD\s+align=middle.*?\>|i);
			chomp($line2);
			$rline .= $line2;
		}
		my @a = split(/(\<\/?B\>|\<\/?FONT.*?\>|\<BR\>|\<\/?P\>|\<I\>|\<P.*?\>)/i, $rline);;
		my $IsPauling = 0;
		my $IsReliable = 0;
		foreach my $s (@a) {
			$IsPauling = 1 if($s =~ /\<I\>/i);
			$IsReliable = 1 if($s =~ /\<B\>/i);
			$s =~ s/^\s+//;
			$s =~ s/\s+$//;
			next if($s eq '' or $s =~ /[\<\>]/);
			if($s =~ /^[\d\.]+$/) { # to A
				$s = $s * 0.01;
				$s = "$s (less reliable)"
					unless($IsReliable);
				$s = "$s (deduced by Pauling(1960) or Ahrens(1952))"
					if($IsPauling);
			}
			push(@radius, $s);
			$IsPauling = 0;
			$IsReliable = 0;
		}
		last;
	}
	close(IN);
	return @radius;
}

sub GetAtomRadius
{
	my ($atnum, $UseApproximation, $Debug) = (@_);
	my ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity) = &GetIonInfFromPeriodicTable2($atnum, $Debug);
	return $AtomicRadius;
}

sub GetIonInfFromPeriodicTable2
{
	my ($atnum, $Debug) = (@_);
	my $AtomJName  = '';
	my $AtomEName  = '';
	my $AtomicMass = '';
	my $FoundYear  = '';
	my $AtomicRadius = '';
	my $MeltingPoint = '';
	my $BoilingPoint = '';
	my $Density      = '';
	my $HeatCapacity = '';
	my $IonizationPotential = '';
	my $ElectronAffinity    = '';
#原子番号 94 
#原子量 239 
#発見年 (年) 1940 
#原子半径 (A) 1.64 
#融点 (oC) 639.5 
#沸点 (oC) 3235 
#密度 (g/cm3) 　 
#比熱 (cal/g oC) 　 
#イオン化エネルギー (eV) 5.8 
#電子親和力 (eV) 

	my $fname = sprintf "A%03dj.html", int($atnum+0.1);
	my $filepath = Deps::MakePath($PeriodicTable2DBDir, $fname, 0);
	unless(open(IN, "<$filepath")) {
		print "Error: $!: $filepath${LF}";# if($Debug);
		return -1;
	}
	while(<IN>) {
		my $line = $_;
#print "line: $line\n";
		if($line =~ m|<title>(.*?)<\/title>|i) {
			$AtomEName = $1;
		}
		elsif($line =~ m|\<font SIZE=\"\+3\">(.*?)(\(.*?\))?\<\/font\>|i) {
			$AtomJName = $1;
		}
		elsif($line =~ m|原子番号|i) {
			$line =~ /\>(\d+)\<\/font/;
		}
		elsif($line =~ m|原子量|i) {
			$line =~ /1\">([\d\.]*)<\/font/;
			$AtomicMass = $1;
		}
		elsif($line =~ m|発見年.*\>(\d*?)\<\/font|i) {
			$FoundYear = $1;
		}
		elsif($line =~ m|原子半径.*\>([\d\.]*).*?\<\/font|i) {
			$AtomicRadius = $1;
		}
		elsif($line =~ m|融点.*\>([-\+\d\.]*)\<\/font|i) {
			$MeltingPoint = $1;
		}
		elsif($line =~ m|沸点.*\>([-\+\d\.]*)\<\/font|i) {
			$BoilingPoint = $1;
		}
		elsif($line =~ m|密度.*\>([\d\.]*)\<\/font|i) {
			$Density = $1;
		}
		elsif($line =~ m|比熱.*\>([\d\.]*)\<\/font|i) {
			$HeatCapacity = $1;
		}
		elsif($line =~ m|イオン化.*\>([-\+\d\.]*)\<\/font|i) {
			$IonizationPotential = $1;
		}
		elsif($line =~ m|電子.*\>([-\+\d\.]*)\<\/font|i) {
			$ElectronAffinity = $1;
		}
	}
	close(IN);
	return ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity);
}

sub GetIonInfFromPeriodicTable1
{
	my ($name, $Debug) = (@_);

	$name = uc $name;
	my $AtmNum = -1;
	my $JName = "";
	my $EName = "";
	my $NameOrigin = "";
#    <TD>番号</TD>
#    <TD>記号</TD>
#    <TD>元素名</TD>
#    <TD>英語名</TD>
#    <TD>名前の由来</TD></TR>

#print "=== Open [$PeriodicTable1DBPath]\n";
	unless(open(IN, "<$PeriodicTable1DBPath")) {
		print "Error: $!: $PeriodicTable1DBPath${LF}";# if($Debug);
		return -1;
	}

	while(<IN>) {
		my $line = $_;
		next if($line !~ m|<TD><A name=.*?>(\d+)<\/A><\/TD>|i);
		$AtmNum = $1;
		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		my $atname   = $1;
#print "at: $AtmNum: [$atname] [$name]\n";
#		next if($name ne uc $atname and ($name =~ /^[\d+-]+$/ and $AtmNum != int($name+0.1)) ) ;
		next if($name ne uc $atname) ;
		next if($name =~ /^[\d\+\-]+$/ and $AtmNum != int($name+0.1));

		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		$JName = $1;
		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		$EName = $1;
		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		$NameOrigin = $1;
		last;
	}
	close(IN);

	return ($AtmNum, $JName, $EName, $NameOrigin);
}

sub GetAtomInformationByKey
{
	my ($atomname, $key, $Debug) = @_;

	my ($AtomicNumber, $IonName, $IonCharge) = &GetAtomInformation($atomname);

	unless(open(IN, "<$PeriodicTable3DB")) {
		print "Can't open $PeriodicTable3DB$LF" if($Debug);
		print "$LF" if($Debug);
		return;
	}
	my $line = <IN>;
	my @DBTitle = split(/,/, $line);
	my @DBValue;
	while(<IN>) {
		@DBValue = split(/,/, $_);
		last if($DBValue[0] == $AtomicNumber);
	}
	close(IN);

	my $i;
	for($i = 0 ; $i < @DBTitle ; $i++) {
		last if($DBTitle[$i] =~ /$key/);
	}
	return $DBValue[$i];
}

sub GetElectronicConfiguraton
{
	my ($nElectron) = @_;
	
	my $CSV = new CSV;
	if(!$CSV->Read($ElectronicConfigurationDB, 0)) {
		return undef;
	}

	my $pLabelArray   = $CSV->LabelArray();
	my $pDataArray    = $CSV->DataArray();
	my $pAtomNumArray = $pDataArray->[2];
	my $nAtomNumArray = @$pAtomNumArray;
	my $i = 0;
	for( ; $i < $nAtomNumArray ; $i++) {
#print "i=$i: $pAtomNumArray->[$i] / $nElectron\n";
		last if($nElectron == $pAtomNumArray->[$i]);
	}
	return undef if($i == $nAtomNumArray);

#print "ne=$nElectron: i=$i\n";
	my (@name, @n);
	for(my $j = 0 ; ; $j++) {
		my $p = $pDataArray->[3+$j];
		last if(!defined $p);
		my $ne = $p->[$i];
		last if(!defined $ne);
		$ne = 0 if($ne eq '');
		$name[$j] = $pLabelArray->[3+$j];
		$n[$j]    = $ne;
#print "$j: $name[$j]: $n[$j]\n";
	}
	return (\@name, \@n);
}

sub SpeculateElectronicConfiguration
{
	my ($nElectron) = @_;

	my $ResCharge = $nElectron;

	my $n1s = 0.0;
	if($ResCharge > 2.0) {
		$n1s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n1s = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n2s = 0.0;
	if($ResCharge > 2.0) {
		$n2s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n2s = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n2p = 0.0;
	if($ResCharge > 6.0) {
		$n2p = 6.0;
		$ResCharge -= 6.0;
	}
	else {
		$n2p = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n3s = 0.0;
	if($ResCharge > 2.0) {
		$n3s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n3s = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n3p = 0.0;
	if($ResCharge > 6.0) {
		$n3p = 6.0;
		$ResCharge -= 6.0;
	}
	else {
		$n3p = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n3d = 0.0;
	my $n4s = 0.0;

	if($ResCharge > 2.0) {
		$n4s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n4s = $ResCharge;
		$ResCharge = 0.0;
	}
	if($ResCharge > 10.0) {
		$n3d = 10.0;
		$ResCharge -= 10.0;
	}
	else {
		$n3d = $ResCharge;
		$ResCharge = 0.0;
	}

	my $n4p = 0.0;
	if($ResCharge > 6.0) {
		$n4p = 6.0;
		$ResCharge -= 6.0;
	}
	else {
		$n4p = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n4d = $ResCharge;

	my @name = qw(1s    2s    2p    3s    3p    3d    4s    4p    4d);
	my @ne   =   ($n1s, $n2s, $n2p, $n3s, $n3p, $n3d, $n4s, $n4p, $n4d);
	return (\@name, \@ne);
}

BEGIN {
}

sub new
{
	my ($module) = @_;
	my $this = {};
	bless $this;
	return $this;
}

sub DESTROY
{
	my $this = shift;
}

sub SetAtomicNumber
{
	my ($this, $AtomicNumber);
	AtomTypeObject::SetAtomicNumber($this, $AtomicNumber);

	&SetAtomInformation();

	return $this->{'AtomNumber'};
}

sub SetAtomName
{
	my ($this,$AtomName) = @_;
	AtomTypeObject::SetAtomName($this, $AtomName);
	my $AtomicNumber      = &GetAtomicNumber($AtomName);
	$this->{AtomicNumber} = $AtomicNumber;

	$this->SetAtomInformation();

	return $AtomName;
}

sub SetAtomInformation
{
	my ($this) = @_;

	my $AtomName = $this->AtomTypeOnly(); #$this->{'AtomName'};
$AtomName =~ s/\{.*?\}//g;
$AtomName =~ s/\d//g;
	my $Charge       = $this->Charge(); #$this->{'Charge'};
	my $AtomicNumber = $this->AtomicNumber(); #$this->{'AtomicNumber'};
	if($AtomicNumber <= 0) {
		($AtomicNumber) = &GetIonInfFromPeriodicTable1($AtomName);
	}
#print "an: [$AtomName] #$AtomicNumber  Charge=$Charge<br>\n";
	my $IonRadius = &GetCationRadius($AtomName, $Charge, 6);
#	if($IonRadius <= 0) {
#		my @rad = &GetIonRadius2($AtomName, $Charge, 6);
#		$IonRadius = $rad[2];
#	}
	my ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity)
			= &GetIonInfFromPeriodicTable2($AtomicNumber);
#	my ($AtmNum, $JName, $EName, $NameOrigin)
#		= &GetIonInfFromPeriodicTable1($AtomicName);
#print "an: $AtomName: $AtomicMass $AtomicNumber\n";

	$this->{AtomicRadius}        = $AtomicRadius;
	$this->{IonRadius}           = $IonRadius;
	$this->{AtomicMass}          = $AtomicMass;
	$this->{IonizationPotential} = $IonizationPotential;
	$this->{ElectronAffinity}    = $ElectronAffinity;
	return 1;
}


"""
