import sys
import shlex
from pprint import pprint


from tklib.tkobject import tkObject
from tklib.tkcrystal.tkcrystalobject import tkCrystalObject
from tklib.tkcrystal.tkcrystal import tkCrystal
from tklib.tkfile import tkFile
from tklib.tkutils import SplitFilePath, BuildCreationDateStr, pint, pfloat
from tklib.tkre import DelQuote
import tklib.tkre as tkre


#=====================================
# CIF root class
#=====================================

class tkCIFObject(tkFile):
    def __init__(self, path = None, mode = 'r', **args):
        self.__crystalname = ''
        self.__samplename  = ''

    def __del__(self):
        pass

    def __str__(self):
        return self.ClassPath()


#=====================================
# CIF data class
#=====================================
    
class tkCIFData(tkCrystalObject):
    def __init__(self, **args):
        super(tkCIFData, self).__init__(**args)
        self.debug = 0
        self.path  = None
# cif data dict
        self.data  = None
# cif input lines list
        self.lines = None
        self.__crystalname = None
        self.__samplename  = None
        self.update(**args)

    def __del__(self):
        super(tkCIFData, self).__del__()

    def __str__(self):
        return self.ClassPath()

    def get(self, key, i = None, defval = None):
#        print(f"get:i={i}  key={key}")
        if type(key) is list:
            for j in range(len(key)):
                ret = self.get(key[j], i, defval = None)
                if ret is not None:
                    return ret
            else:
                return defval

        if type(key) is int:
            try:
                return self.lines[key]
            except:
                return defval

        if key == 'path':
            return self.path

        key1 = key
        if i is not None and i != '':
            key1 = key + f'[{i}]'
#        print("key1=", key1)
#        for k in  self.data.keys():
#            print("   {}:{}".format(k, self.data[k]))
        try:
            val = self.data[key1]
            return val
        except:
            return defval


    def geterrorf(self, key, i = None, defval = None):
        val = self.get(key, i, defval)
        if val is None:
            return defval
#        print("geterrorf: {} = {}: type(val)={}".format(key, val, type(val)))
        try:
            val = tkre.Sub(r'\(.*\)', '', val)
        except:
            pass
#        print("           {} = {}".format(key, val))
        try:
            return pfloat(val)
        except:
           return defval

    def getf(self, key, i = None, defval = None):
        val = self.get(key, i, defval)
        try:
            return pfloat(val)
        except:
           return defval

    def geti(self, key, i = None, defval = None):
        val = self.get(key, i, defval)
        try:
            return pint(val)
        except:
            pass
        try:
            return int(pfloat(val))
        except:
            return defval

    def CheckCIF1(self, key):
        if self.geterrorf(key) is None:
            print("Error in tkcrystal.tkcif.tkCIFData.CheckCIF1: key [{}] is not found in [{}]"
                    .format(key, self.path))
            return 0
        return 1

    def CheckCIF(self):
        if self.CheckCIF1('_cell_length_a')             \
            and self.CheckCIF1('_cell_length_b')        \
            and self.CheckCIF1('_cell_length_c')        \
            and self.CheckCIF1('_cell_angle_alpha')     \
            and self.CheckCIF1('_cell_angle_beta')      \
            and self.CheckCIF1('_cell_angle_gamma')     \
            and self.CheckCIF1('_atom_site_fract_x[0]') \
            and self.CheckCIF1('_atom_site_fract_y[0]') \
            and self.CheckCIF1('_atom_site_fract_z[0]'):
            return 1
        exit()

    def LatticeParameters(self):
        latt = self.geterrorf('_cell_length_a'), \
               self.geterrorf('_cell_length_b'), \
               self.geterrorf('_cell_length_c'), \
               self.geterrorf('_cell_angle_alpha'), \
               self.geterrorf('_cell_angle_beta'), \
               self.geterrorf('_cell_angle_gamma')
                       
        return latt


    def dataprint(self, key, i = None):
        if type(key) is not list:
            key = [key]

        for j in range(len(key)):
             ret = self.get(key[j], i = i, defval = None)
             if ret is not None:
                 print("{}: {}".format(key[j], self.get(key[j], i)))
                 return 1
        print("Error in {}.{}: Invalid key {} for tkCIFData.data"
              .format(self.ClassPath(), sys._getframe().f_code.co_name, key))
        return 0

    def InsertSpaceToSPGName(self, SPGName):
        if tkre.Search(r'[RH]$', SPGName) and tkre.Search(r'\s.+\S\s+[RHS]$', SPGName):
            return SPGName
        elif tkre.Search(r'\s', SPGName):
            return SPGName

        tkre.Sub(r'(.)', r'\1', SPGName)
        tkre.Sub(r'\s\s+', r' ', SPGName)
        tkre.Sub(r'- ', r'-', SPGName)
        tkre.Sub(r' \/ ', r'\/', SPGName)
        tkre.Sub(r'([^\-])2 1', r'\{1}21', SPGName)
        tkre.Sub(r'([^\-])3 1', r'\{1}31', SPGName)
        tkre.Sub(r'([^\-])3 2', r'\{1}32', SPGName)
        tkre.Sub(r'([^\-])4 1', r'\{1}41', SPGName)
        tkre.Sub(r'([^\-])4 2', r'\{1}42', SPGName)
        tkre.Sub(r'([^\-])4 3', r'\{1}43', SPGName)
        tkre.Sub(r'([^\-])6 1', r'\{1}61', SPGName)
        tkre.Sub(r'([^\-])6 2', r'\{1}62', SPGName)
        tkre.Sub(r'([^\-])6 3', r'\{1}63', SPGName)
        tkre.Sub(r'([^\-])6 4', r'\{1}64', SPGName)
        tkre.Sub(r'([^\-])6 5', r'\{1}65', SPGName)

        return SPGName

    def FillCIFData(self):
        crystal = tkCrystal()
        crystal.SetCrystalName(self.CrystalName())
#        crystal->SetLatticeParameters(self.LatticeParameters())
#        self.SetContent("_cell_volume", crystal.Volume())
#        SPG = self.SpaceGroup()
    """
	$crystal->SetCSpaceGroup($SPG);
	$crystal->SetCAtomTypeList($this->GetCAtomTypeList() );
	$crystal->SetCAsymmetricAtomSiteList($this->GetCAsymmetricAtomSiteList() );
	$crystal->ExpandCoordinates();

	my @MultList = $crystal->GetCnMultiplicityExpandedAtomSiteList();
	for(my $i = 0 ; $i < @MultList ; $i++) {
		$this->SetContent("_atom_site_symmetry_multiplicity", 
				$i+1, $MultList[$i]);
	}

	if($this->GetContent("_symmetry_cell_setting") eq '') {
		$this->SetContent("_symmetry_cell_setting", 
				$SPG->LatticeSystem());
	}
	if($SPG->SPGName() =~ /^P\s*1$/i) {
		$this->SetContent("_symmetry_cell_setting", 'triclinic');
	}

	my %AtomCount;
	my $nAsymmetricAtomSite = $crystal->nAsymmetricAtomSite();
	for(my $i = 0 ; $i < $nAsymmetricAtomSite ; $i++) {
		my $site = $crystal->GetCAtomSite($i);
		my $mult = $MultList[$i];
		my $name = $site->AtomNameOnly();
		$AtomCount{$name} += $mult * $site->Occupancy();
#print "occ: ", $site->Occupancy(), "<BR>";
	}
	my $s = '';
	my @keys = keys %AtomCount;
	for(my $i = 0 ; $i < @keys ; $i++) {
		my $name = $keys[$i];
		my $n    = $AtomCount{$name};
		my $s1 = substr("$n", 0, 5);
#print "$n:$s1<br>";
		$s = "$s $name$s1";
	}
	$s =~ s/^\s+//;
	$this->SetContent("_chemical_formula_structural", $s);
	$this->SetContent("_chemical_formula_sum", $s);
	$s =~ s/\s//g;
	$this->SetContent("_chemical_name_systematic", $s);
	$this->SetContent("_cell_formula_units_Z", 1);
	
	my $date = $this->GetContent("_audit_creation_date");
	if($date eq '') {
		my ($sec, $min, $hour, $mday, $mon, $year,
		        $wday, $yday, $isdst) = localtime(time());
		$mon++;
		$year += 1900;
		my $s = sprintf "%04d/%02d/%02d", $year, $mon, $mday;
		$this->SetContent("_audit_creation_date", $s);
	}
	
	$this->SetContent("_audit_creation_method", "generated by Perl:CIF.pm 0.1");
	return $crystal;
}
    """

    def GetCrystal(self):
        cry = tkCrystal()
        cifdata = self.data

        cry.path = self.path
        dir, filename, filebody, ext = SplitFilePath(cry.path)

        cry.ChemicalName    = self.get('_chemical_name_systematic',     defval = '')
        cry.ChemicalFormula = self.get('_chemical_formula_structural',  defval = '')
        cry.SetVolume(self.geterrorf('_cell_volume', defval = None))
        cry.ChemicalFormulaUnit = self.get('_cell_formula_units_Z',           defval = None)

        spg_name = self.get(
                ['_symmetry_space_group_name_H-M', '_space_group_name_H-M_alt'],
                defval = '')
        ispg = self.geti(
                ['_symmetry_Int_Tables_number', '_space_group_IT_number'],
                defval = '')
        cry.SetSpaceGroup(name = spg_name, number = ispg)
        cry.SetLatticeParameters(self.LatticeParameters(), GuessLatticeSystem = 0)

        if cry.ChemicalFormula != '':
            cry.SetCrystalName(cry.ChemicalFormula)
            cry.SetSampleName(cry.ChemicalName)
        else:
            cry.SetSampleName(filebody)

        c = 0
        while 1:
            sym = self.get(
                    ['_symmetry_equiv_pos_as_xyz', '_space_group_symop_operation_xyz'], 
                    i = c, defval = None)
            if sym:
#                print("add sym:", sym, "  c=", c)
                cry.AddSymmetry(sym)
                c += 1
            else:
                break
        
        c = 0
        while 1:
            atomtype = self.get('_atom_type_symbol', c)
            if(atomtype):
                charge = self.getf('_atom_type_oxidation_number', c, defval = 0.0)
                cry.AddAtomType(atomtype, charge)
                c += 1
            else:
                break

        c = 0
        while 1:
            label = self.get('_atom_site_label', c)
            if(label):
                name = self.get('_atom_site_type_symbol',            c, defval = None)
                m    = self.geti('_atom_site_symmetry_multiplicity', c, defval = 1)
                ws   = self.get('_atom_site_Wyckoff_symbol',         c, defval = '')
                x    = self.geterrorf('_atom_site_fract_x',          c, defval = None)
                y    = self.geterrorf('_atom_site_fract_y',          c, defval = None)
                z    = self.geterrorf('_atom_site_fract_z',          c, defval = None)
                occ  = self.geterrorf('_atom_site_occupancy',        c, defval = 1.0)
                h    = self.geti('_atom_site_attached_hydrogens',    c, defval = 0)
                if name is None:
                    name = self.get('_atom_site_label',            c, defval = None)
                    list = tkre.Match(r'([A-Z][a-z]?)', name)

#                iAtomType, AtomType = cry.AddAtomType(name)
                cry.AddAtomSite(label = label, name = name, m = m, ws = ws,
                                pos = [x, y, z], occ = occ, hydrogen = h)
                c += 1
            else:
                break

        cry.ExpandCoordinates()

        return cry


    def Print(self, PrintLines = 0):
        self.print(PrintLines = PrintLines)
        
    def print(self, PrintLines = 0):
        cifdata = self.data

        print("====== tkcifobject.Print start ======") 
        print("path: {}".format(self.path))
        self.dataprint('_chemical_name_systematic')
        self.dataprint('_chemical_formula_structural')

        print("\nUnit cell:")
        latt = self.LatticeParameters()
        print('latt: ', latt)
        self.dataprint('_cell_volume')
        self.dataprint('_cell_formula_units_Z')

        print("\nSpace group:")
        self.dataprint(['_symmetry_space_group_name_H-M', '_space_group_name_H-M_alt'])
        self.dataprint(['_symmetry_Int_Tables_number', '_space_group_IT_number'])

        print("\nSymmetry:")
        c = 0
        while 1:
            idx = self.geti('_symmetry_equiv_pos_site_id', i = c, defval = c)
            sym = self.get(
                    ['_symmetry_equiv_pos_as_xyz', '_space_group_symop_operation_xyz'], 
                    i = idx, defval = None)
#            print("idx,sym=", idx, sym)                    
            if sym:
                print("{}: {}".format(idx, sym))
                c += 1
            else:
                break

        print("\nAtom Type:")
        c = 0
        while 1:
            atomtype = self.get('_atom_type_symbol', c)
            if(atomtype):
                charge = self.getf('_atom_type_oxidation_number', c, defval = 0.0)
                print("{}: {}".format(atomtype, charge))
                c += 1
            else:
                break

        print("\nAtom Site:")
        c = 0
#        for key in self.data.keys():
#            print(" {}={}".format(key, self.data[key]))
        while 1:
            label = self.get('_atom_site_type_symbol', c, defval = None)
            if label is None:
                label = self.get('_atom_site_label', c, defval = None)
            if label:
                name = self.get('_atom_site_type_symbol',            c, defval = None)
                m    = self.geti('_atom_site_symmetry_multiplicity', c, defval = 1)
                ws   = self.get('_atom_site_Wyckoff_symbol',         c, defval = '')
                x    = self.geterrorf('_atom_site_fract_x',          c, defval = None)
                y    = self.geterrorf('_atom_site_fract_y',          c, defval = None)
                z    = self.geterrorf('_atom_site_fract_z',          c, defval = None)
                occ  = self.geterrorf('_atom_site_occupancy',        c, defval = 1.0)
                h    = self.geti('_atom_site_attached_hydrogens',    c, defval = 0)
                if name is None:
                    name = self.get('_atom_site_label',            c, defval = None)
                    list = tkre.Match(r'([A-Z][a-z]?)', name)
                    try:
                        name = list[1]
                    except:
                        pass
                if name is None or x is None or y is None or z is None:
                    print("Error: tkcrystal.tkcifobject.Print: name ({}) or (x, y, z) ({}, {}, {}) is None)"
                        .format(name, x, y, z))
                    exit(0)
                print("#{:04d}: {:4} {:4} ({:8.4f}, {:8.4f}, {:8.4f}) occ={:8.4f}  ws={}  m={}"
                        .format(c, label, name, x, y, z, occ, ws, m))
                c += 1
            else:
                break

        if PrintLines:
            print("\nInput data:")
            c = 0
            while 1:
                ret = self.get(c)
                if(ret):
                    print(ret.strip())
                    c += 1
                else:
                    break

        print("====== tkcifobject.Print end ======") 

    def nSymmetryOperation(self):
        n = 0
        while 1:
            val = self.get('_symmetry_equiv_pos_as_xyz', i = n, defval = None)
            if val is None:
                return n
            n += 1

    def SymmetryOperation(self, i):
        val = self.get('_symmetry_equiv_pos_as_xyz', i = i, defval = None)
        if val is None:
            return val
        return DelQuoate(val)

    def WriteKeyVal(self, out, key, defval = None, i = None):
        val = self.get(key, i = i, defval = defval)
#        print("\nlines:\n")
#        for key in self.data.keys():
#            print("  {}: {}".format(key, self.data[key]))
#        print("  key=", key, "i=", i, "  val=", val)
        if val is None:
            return 0
        out.Write("%-35s%s\n" % (key, val))
        return 1

    def WriteSimpleCIFFile(self, NewFile, WritePublication = 1):
        self.FillCIFData()
#ファイル名を（ベース名, ディレクトリ名, 拡張子）に分解
#        dirname, basename, filebody, ext = SplitFilePath(NewFile)

        out = tkFile(NewFile, 'w')
        if out is None:
            return 0
#        binmode(OUT) if(defined $strCRLF and $strCRLF =~ /unix/i);

        val = self.get('_database_code_ICSD', defval = 99999)
        out.Write("data_{}-ICSD\n".format(val))
        out.Write("\n")
        self.WriteKeyVal(out, '_audit_creation_date',         None)
        self.WriteKeyVal(out, '_audit_creation_method',       'python tkcifobject.py 0.1')
        self.WriteKeyVal(out, '_chemical_name_systematic',    'undefined')
        self.WriteKeyVal(out, '_chemical_formula_structural', 'undefined')
        self.WriteKeyVal(out, '_chemical_formula_sum',        'undefined')
        if WritePublication:
            self.WriteKeyVal(out, '_journal_name_full',        'undefined')
            out.Write("\n")

        self.WriteKeyVal(out, '_cell_length_a',        '')
        self.WriteKeyVal(out, '_cell_length_b',        '')
        self.WriteKeyVal(out, '_cell_length_c',        '')
        self.WriteKeyVal(out, '_cell_angle_alpha',     '')
        self.WriteKeyVal(out, '_cell_angle_beta',      '')
        self.WriteKeyVal(out, '_cell_angle_gamma',     '')
        self.WriteKeyVal(out, '_cell_volume',          None)
        self.WriteKeyVal(out, '_cell_formula_units_Z', None)
        out.Write("\n")

        self.WriteKeyVal(out, '_symmetry_space_group_name_H', None)
        self.WriteKeyVal(out, '_symmetry_Int_Tables_number', None)
        self.WriteKeyVal(out, '_symmetry_cell_setting', None)
        out.Write("loop_\n")
        out.Write('_symmetry_equiv_pos_as_xyz\n')
        nsymop = self.nSymmetryOperation()
        for i in range(nsymop):
            val = self.get('_symmetry_equiv_pos_as_xyz', i = i, defval = '')
            val = DelQuote(val)
            out.Write("  '{}'\n".format(val))
        out.Write("\n")

        out.Write("loop_\n")
        out.Write("_atom_type_symbol\n")
        out.Write("_atom_type_oxidation_number\n")
        c = 0
        while 1:
            name = self.get('_atom_type_symbol', i = c, defval = None)
            if name is None:
                break
            charge = self.get('_atom_type_oxidation_number', i = c, defval = 0.0)
            out.Write("  %-6s%6s\n" % (name, charge))
            c += 1
        out.Write("\n")

        out.Write("loop_\n")
        out.Write("_atom_site_label\n")
        out.Write("_atom_site_type_symbol\n")
        out.Write("_atom_site_symmetry_multiplicity\n")
        out.Write("_atom_site_Wyckoff_symbol\n")
        out.Write("_atom_site_fract_x\n")
        out.Write("_atom_site_fract_y\n")
        out.Write("_atom_site_fract_z\n")
        out.Write("_atom_site_occupancy\n")
        c = 0
        while 1:
            label = self.get('_atom_site_label',       i = c, defval = None)
            type  = self.get('_atom_site_type_symbol', i = c, defval = None)
            if type is None:
                break
            m   = self.get('_atom_site_symmetry_multiplicity', i = c, defval = 0.0)
            ws  = self.get('_atom_site_Wyckoff_symbol',        i = c, defval = 0.0)
            x   = self.get('_atom_site_fract_x',               i = c, defval = 0.0)
            y   = self.get('_atom_site_fract_y',               i = c, defval = 0.0)
            z   = self.get('_atom_site_fract_z',               i = c, defval = 0.0)
            occ = self.get('_atom_site_occupancy',             i = c, defval = 0.0)
            out.Write("  %-6s %-6s %2s %s   %s %s %s  %s\n" 
                    % (label, type, m, ws, x, y, z, occ))
            c += 1
        out.Write("\n")
        out.Write("\n")
        out.Write("#End of data_${dbcodeICSD}-ICSD\n")
        out.Write("\n")
        out.Write("\n")

        out.Close()

    def CreateCIFFileFromCCrystal(self, cry, NewFile, IsChooseRandomly = 0, IsPrint = 1):
        out = tkFile(NewFile, 'w')
        if out is None:
            return 0
#       binmode(OUT) if($strCRLF =~ /unix/i);

        val =  99999;
        out.Write("data_{}\n".format(val))
        out.Write("\n")

        out.Write("%-35s%s\n" % ("_audit_creation_date", BuildCreationDateStr()))
        out.Write("%-35s'%s'\n" % ("_audit_creation_method", 'python tkcifobject.py 0.1'))
        out.Write("\n")
        Composition = cry.ChemicalComposition()
        Composition = tkre.Sub(r'\s', '', Composition)
        name = Composition
        if name == '':
            name = cry.CrystalName()
        if name == '':
            name = cry.SampleName()
        if name == '':
            name = 'undefined'
        out.Write("%-35s'%s'\n" % ("_chemical_formula_structural", cry.ChemicalComposition()))
        out.Write("%-35s'%s'\n" % ("_chemical_formula_sum", cry.SumChemicalComposition()))
        latt = cry.LatticeParameters()
        out.Write("\n")

        out.Write("%-35s %s\n" % ("_cell_length_a",    latt[0]))
        out.Write("%-35s %s\n" % ("_cell_length_b",    latt[1]))
        out.Write("%-35s %s\n" % ("_cell_length_c",    latt[2]))
        out.Write("%-35s %s\n" % ("_cell_angle_alpha", latt[3]))
        out.Write("%-35s %s\n" % ("_cell_angle_beta",  latt[4]))
        out.Write("%-35s %s\n" % ("_cell_angle_gamma", latt[5]))
        out.Write("\n")

        SPG = cry.SpaceGroup()
        op = SPG.GetSymmetryOperationList()
        out.Write("%-35s %s\n" % ("_symmetry_Int_Tables_number", SPG.iSPG()))
        out.Write("\n")

        out.Write("loop_\n")
        out.Write("_symmetry_equiv_pos_as_xyz\n")
        for i in range(len(op)):
            out.Write("  {}\n".format(op[i]))
        out.Write("\n")

        at = cry.AtomTypeList()
        out.Write("loop_\n")
        out.Write("_atom_type_symbol\n")
        out.Write("_atom_type_oxidation_number\n")
        for i in range(len(at)):
            atype  = at[i].AtomTypeOnly()
            charge = at[i].Charge()
            out.Write("  %-2s  %f\n" % (atype, charge))
        out.Write("\n")

        ast = cry.AtomSiteList()
        out.Write("loop_\n")
        out.Write("_atom_site_label\n")
        out.Write("_atom_site_type_symbol\n")
        out.Write("_atom_site_fract_x\n")
        out.Write("_atom_site_fract_y\n")
        out.Write("_atom_site_fract_z\n")
        out.Write("_atom_site_occupancy\n")
        for i in range(len(ast)):
            label = ast[i].Label()
            name  = ast[i].AtomNameOnly()
            pos   = ast[i].Position()
            occ   = ast[i].Occupancy()
            out.Write("  %-6s %-2s  %10.6f %10.6f %10.6f  %f\n" % 
                (label, name, pos[0], pos[1], pos[2], occ))

        out.Write("\n")
        out.Write("#End of data_99999-ICSD\n")
        out.Close()


"""
	$Composition = $Crystal->ChemicalComposition();
	printf OUT "%-35s'%s'\n", "_chemical_formula_structural", $Composition;
	my $SumComposition = $Crystal->SumChemicalComposition();
	printf OUT "%-35s'%s'\n", "_chemical_formula_sum",  $SumComposition;
	print OUT "\n";

	my ($a,$b,$c,$alpha,$beta,$gamma) = $Crystal->LatticeParametersByOutputMode(0);
#	$a = sprintf("%12.6f", $a);
#	$b = sprintf("%12.6f", $b);
#	$c = sprintf("%12.6f", $c);
#	$alpha = sprintf("%12.6f", $alpha);
#	$beta  = sprintf("%12.6f", $beta);
#	$gamma = sprintf("%12.6f", $gamma);
	$a = Utils::Round($a, 6);
	$b = Utils::Round($b, 6);
	$c = Utils::Round($c, 6);
	$alpha = Utils::Round($alpha, 6);
	$beta  = Utils::Round($beta, 6);
	$gamma = Utils::Round($gamma, 6);
	printf OUT "%-35s%s\n", "_cell_length_a", $a;
	printf OUT "%-35s%s\n", "_cell_length_b", $b;
	printf OUT "%-35s%s\n", "_cell_length_c", $c;
	printf OUT "%-35s%s\n", "_cell_angle_alpha", $alpha;
	printf OUT "%-35s%s\n", "_cell_angle_beta",  $beta;
	printf OUT "%-35s%s\n", "_cell_angle_gamma", $gamma;
	printf OUT "%-35s%s\n", "_cell_volume", $Crystal->Volume();
	printf OUT "%-35s%s\n", "_cell_formula_units_Z", $Crystal->FormulaUnit();
	print OUT "\n";

	my $SPGName = $Crystal->SPGNameByOutputMode();
	$SPGName = $this->InsertSpaceToSPGName($SPGName);
#print "SPGName: $SPGName<br>\n";
	my $iSPG    = $Crystal->iSPGByOutputMode();
	my $SPG = $Crystal->GetCSpaceGroup();
	my $LatticeSystem = $SPG->LatticeSystem();
	printf OUT "%-35s'%s'\n", "_symmetry_space_group_name_H-M", $SPGName;
	printf OUT "%-35s%s\n", "_symmetry_Int_Tables_number", $iSPG;
	printf OUT "%-35s%s\n", "_symmetry_cell_setting", $LatticeSystem;
	print OUT "\n";

	my @SymmetryOperation = $SPG->ExpandSymmetryOperation();
	my $nsymop = @SymmetryOperation;
#print "nsymop: $nsymop<br>\n";
	if($iSPG == 1 or $nsymop == 0) {
#		$nsymop = 1;
#		$SymmetryOperation[1] = "x,y,z";
		$nsymop = 0;
		$SymmetryOperation[0] = "x,y,z";
	}
	print OUT "loop_\n";
	printf OUT "%-35s%s\n", "_symmetry_equiv_pos_as_xyz", "";
		print OUT "  'x,y,z'\n";
	for(my $i = 0 ; $i <= $nsymop ; $i++) {
		my $SymOp = $SymmetryOperation[$i];
		Utils::DelSpace($SymOp);
		next if(!defined $SymOp or $SymOp =~ /^x\s*,\s*y\s*,\s*z$/i);
		print OUT "  \'$SymOp\'\n" if($SymOp ne '');
	}
	print OUT "\n";

	my @AtomTypeList = $Crystal->GetCAtomTypeList();
	print OUT "loop_\n";
	printf OUT "%-35s%s\n", "_atom_type_symbol", "";
	printf OUT "%-35s%s\n", "_atom_type_oxidation_number", "";
	for(my $i = 0 ; $i < @AtomTypeList ; $i++) {
		my $atom = $AtomTypeList[$i];
		my $atomtype = $atom->AtomNameOnly();
next if($atomtype eq '');
		my $oxidationstate = $atom->Charge();
		$oxidationstate = 0 unless(defined $oxidationstate);
		printf OUT "  %-6s%6s\n", $atomtype, $oxidationstate;
	}
	print OUT "\n";

	my @ExpandedAtomSiteList = $Crystal->GetCExpandedAtomSiteListByOutputMode();
	my $nExpandedAtomSite    = @ExpandedAtomSiteList;
	print OUT "loop_\n";
	print OUT "_atom_site_label\n";
	print OUT "_atom_site_type_symbol\n";
	print OUT "_atom_site_symmetry_multiplicity\n";
	print OUT "_atom_site_Wyckoff_symbol\n";
	print OUT "_atom_site_fract_x\n";
	print OUT "_atom_site_fract_y\n";
	print OUT "_atom_site_fract_z\n";
	print OUT "_atom_site_occupancy\n";
	print OUT "_atom_site_attached_hydrogens\n";
	print OUT "_atom_site_calc_flag\n";
#Occupancyが1のサイト
	my $count = 0;
	my %AtomCount;
	for(my $i = 0 ; $i < $nExpandedAtomSite ; $i++) {
		my $atom      = $ExpandedAtomSiteList[$i];
		my $atomname  = $atom->AtomNameOnly();
		my $charge    = $atom->Charge();
		my ($x,$y,$z) = $atom->Position(1);
		my $occupancy = $atom->Occupancy();
		my $mult      = $atom->Multiplicity();
		$mult = 1 if($Crystal->OutputMode() =~ /expand/i or $IsChooseRandomly);
		$mult = 1 if($mult <= 0.0);
		next if($occupancy < 0.9999);

		$AtomCount{$atomname}++;
		$count++;
		my $label = $atomname . $AtomCount{$atomname};
		printf OUT "  %-8s%-8s%2d %s %11.6f %11.6f %11.6f %9.4f  %s %s\n",
			$label, $atomname, $mult, "a", $x, $y, $z,
			$occupancy, "0", "a";
	}
#Occupancyが1未満のサイト
	if(@ExpandedAtomSiteList > $count and !$IsChooseRandomly) {
my $PrintShareAtomWarning = 0;
		if($PrintShareAtomWarning) {
			print "<H3>***Warning***: One line should be removed from $NewFile.</H3>\n";
			print OUT "*** Shared atoms: Remove this line for calculation: OCC=";
			for(my $i = 0 ; $i < @ExpandedAtomSiteList ; $i++) {
				my $atom      = $ExpandedAtomSiteList[$i];
				my $occupancy = $atom->Occupancy();
				next if($occupancy >= 0.9999);
				printf OUT "%lg ", $occupancy;
			}
			print OUT "\n";
		}
	}
	for(my $i = 0 ; $i < $nExpandedAtomSite ; $i++) {
		my $atom      = $ExpandedAtomSiteList[$i];
		my $atomname  = $atom->AtomNameOnly();
		my $charge    = $atom->Charge();
		my ($x,$y,$z) = $atom->Position(1);
		my $occupancy = $atom->Occupancy();
		my $mult      = $atom->Multiplicity();
		$mult = 1 if($Crystal->OutputMode() =~ /expand/i or $IsChooseRandomly);
		next if($occupancy >= 0.9999);

		$AtomCount{$atomname}++;
		$count++;
		$occupancy = 1 if($IsChooseRandomly);
		my $label = $atomname . $AtomCount{$atomname};
		printf OUT "  %-8s%-8s%2d %s %11.6f %11.6f %11.6f %9.4f  %s %s\n",
			$label, $atomname, $mult, "a", $x, $y, $z,
			$occupancy, "0", "a";
	}

	print OUT "\n";
	print OUT "\n";
	print OUT "#End of data_${dbcodeICSD}-ICSD\n";
	print OUT "\n";
	print OUT "\n";

	close(OUT);

	return 1;
}
"""

"""

#============================================================
#　コンストラクタ、デストラクタ
#============================================================
sub new
{
	my ($module, $filename) = @_;
	my $this = {};
	bless $this;
	$this->SetFileName($filename);
	return $this;
}

sub DESTROY
{
	my $this = shift;
}

#============================================================
#　一般関数
#============================================================
sub ReadCrystal
{
	my ($this, $cifpath) = @_;

	unless($this->Read($cifpath, 0)) {
		return undef;
	}
	return $this->GetCCrystal();
}

sub SetCrystalName
{
	my ($this,$name) = @_;
	Utils::DelSpace($name);
	$this->{'CrystalName'} = $name;
	return $name;
}

sub CrystalName
{
	my ($this) = @_;
	return $this->{'CrystalName'};
}

sub SetFormula
{
	my ($this,$s) = @_;
	$this->{'_chemical_formula_sum'} = $s;
	$this->{'_chemical_formula_structural'} = $s
		if($this->{'_chemical_formula_structural'} eq '');
	return $s;
}

sub Formula
{
	my ($this) = @_;
	return $this->{'_chemical_formula_structural'} if(defined $this->{'_chemical_formula_sum'});
	return $this->{'_chemical_formula_sum'} if(defined $this->{'_chemical_formula_sum'});
	return undef;
}
sub SetFileName
{
	my $this = shift;
	my $s = shift;
	Utils::DelSpace($s);
#no strict;
	$this->{'FileName'} = $s;
#use strict;	
}

sub FileName
{
	my $this = shift;
	return $this->{'FileName'};
}

sub Read
{
	my ($this, $f) = @_;
#	$this->SetFileName($f) if($f ne '');

	my $in = new JFile;
	unless($in->Open($f, "r")) {
#		print "$!: $f\n";
		return 0;
	}
	my $ret = $this->ReadByJFile($in);
	$in->Close();

	return $ret;
}

sub ReadByJFile
{
	my ($this, $in) = @_;

	$this->SetFileName($in->{FileName});
	my $FileEnd = 0;
#最初の一行
	my $header;
	while(!$FileEnd) {
		$header = $in->ReadLine();
		if(!defined $header) {
			return 0;
		}
		Utils::DelSpace($header);
		if($header =~ /^_/) {
			$in->rewind();
			last;
		}
		if($header ne '') {
			$this->{header} = $header;
			last;
		}
	}

	my $ExitLoop = 0;
	while(!$FileEnd) {
		last if($ExitLoop);

		my $line = $in->ReadLine();
		last if(!defined $line);

		Utils::DelSpace($line);
if($Debug) {
	print "line: $line$LF";
}
# 空行はスキップ
		next if($line =~ /^$/);
#		next if($line =~ /^#/);
# #Endを見つけたら終わる
		if($line =~ /^#End/i) {
			$FileEnd = 1;
			last;
		}
		next if($line =~ /^#/);
#print "l0: $line\n";

#loop_の無いデータを読み込み
		if($line =~ /^_/) {
			my ($key, $content) = ($line =~ /^(\S+?)\s+(.*)$/);
#			if(defined $key and $key ne '') {
			if($content ne '') {
				$this->{$key} = $content;
				next;
			}
			else {
				$line = $in->ReadLine();
#print "ll0: $line";
				Utils::DelSpace($line);
				next if($line ne ';');
				$content = '';
				while(1) {
					$line = $in->ReadLine();
#print "ll1: $line";
					Utils::DelSpace($line);
					last if($line eq ';' or $line =~ /^#End/i);
					$content .= "$line\n";
				}
				$content =~ s/[\r\n]+$//;
#print "cont: [$content]\n";
				$this->{$key} = $content;
				next;
			}
		}
#loop_のデータを読み込み
		elsif($line =~ /^loop_$/i) {
#print "l: $line\n";
			my @Params;
			my $Count = 0;
			my $line2 = '';
#loop_内のパラメータ名を取得
#			my $pos = 0;
			my $pos = $in->tell();
#print "loop_<br>\n";
			while(!$FileEnd) {
				$line2 = $in->ReadLine();
#print "l2: $line2";
				last if(!defined $line2);
				if($line2 =~ /^#End/i) {
					$FileEnd = 1;
					last;
				}
				next if($line2 =~ /^#/);

				Utils::DelSpace($line2);
#print "line2: $line2<br>\n";
				unless($line2 =~ /^(_|loop_)/) {
#2006/5/4 Na.cifのためにコメントをはずす
					$in->seek($pos, 0) if($pos > 0);
					last;
				}

				$Params[$Count] = $line2;
				$Count++;
#2006/5/4 Na.cifのためにコメントをはずす
				$pos = $in->tell();
			}
#$Debug=1;
if($Debug) {
	foreach my $s (@Params) {
		print "Key: $s$LF";
	}
}
#loop_内のパラメータを取得
#2006/5/4 Na.cifのために''で初期化するように変更
#			my $LoopContent = $line2;
			my $LoopContent = '';
			$pos = 0;
			while(!$FileEnd) {
				$line2 = $in->ReadLine();
				last if(!defined $line2);
				if($line2 =~ /^#End/i) {
					$FileEnd = 1;
					last;
				}
				next if($line2 =~ /^#/);
#print "l3: $line2";

				Utils::DelSpace($line2);
				next if($line2 eq '');
#				last if($line2 eq '');
#次のloop_を見つけたら、LoopContentを解析
				if($line2 =~ /^(_|loop_)/) {
					$in->seek($pos, 0) if($pos > 0);
					last;
				}
				$LoopContent = "$LoopContent $line2";
				$pos = $in->tell();
			}
			last if($ExitLoop);

#print "LoopC: [$LoopContent]\n";
			my $pt = 0;
			my $c = 0;
			while($LoopContent =~ /(\'.*?\'|\S+)/g) {
				my $s = $1;
				my $i = int($c / $Count) + 1;
				$c++;
#				if($ShowCIFInf ne '') {
#					print "  <b>$Params[$pt]" . "[$i] " . ":</b> $s\n";
#					print "  $i: $Params[$pt]: $s\n";
#				}

#				$Params[$pt] =~ s/[\r\n\s]//g;
				my $key = "${Params[$pt]}[$i]";
#print "k: $key: $s\n";
				$this->{$key} = $s;
if($Debug) {
	print "akey: {$key}  :  $s$LF";
}
				$pt++;
				$pt = 0 if($pt >= $Count);
			}
		}
		last if($ExitLoop);
	}
	$this->SetOtherParameters();
#exit;

	return 1;
}

sub SetOtherParameters
{
	my ($this) = @_;
	my $nAtomType = 0;
	for(my $i = 1 ; ; $i++) {
		my $val = &GetValue($this, "_atom_type_symbol", $i);
#		if(!defined $val or $val eq '') {
#			$val = &GetValue($this, "_atom_site_label", $i);
#			$val =~ s/[\d+-]+//g;
#		}
		last if(!defined $val or $val eq '');
		$this->{"iAtomType_$val"} = $i;
		$nAtomType++;
	}
	$this->{"nAtomType"} = $nAtomType;

	my $nAsymmetricAtomSite = 0;
	for(my $i = 1 ; ; $i++) {
		my $val = &GetValue($this, "_atom_site_fract_x", $i);
		last if(!defined $val or $val eq '');
		$nAsymmetricAtomSite++;
	}
	$this->{"nAsymmetricAtomSite"} = $nAsymmetricAtomSite;

#	if($nAtomType == 0) {
		for(my $i = 1 ; $i <= $nAsymmetricAtomSite ; $i++) {
			my $val = &GetValue($this, "_atom_site_type_symbol", $i);
			if(!defined $val or $val eq '') {
				$val = &GetValue($this, "_atom_site_label", $i);
				$val =~ s/[\d+-]+//g;
			}
#print "v: $val$LF";
			if($val =~ /^(\w+)([\+-])([\d\.]+)$/i) {
				if($3 == 1) {
					$val = "$1$2";
				}
				else {
					$val = "$1$3$2";
				}
#print " => v: $val$LF";
			}
#print "v: $val$LF";
			$this->AddAtomType($val);
			&SetContent($this, "_atom_site_type_symbol", $i, $val);
		}
#	}

	my $a = &GetValue($this, "_cell_length_a");
	my $b = &GetValue($this, "_cell_length_b");
	my $c = &GetValue($this, "_cell_length_c");
	my $alpha = &GetValue($this, "_cell_angle_alpha");
	my $beta  = &GetValue($this, "_cell_angle_beta");
	my $gamma = &GetValue($this, "_cell_angle_gamma");
	my $lattice = $this->{"_symmetry_cell_setting"};
#print "lat: $a $b $c$LF";

	my $SPG = new SpaceGroup();
	my ($SPGName, $iSPG) = $this->GetSpaceGroup();
	$SPG->SetSPGName($SPGName);
	$SPG->SetiSPG($iSPG);
#print "SPGName1 in CIFObject: $SPGName\n";
	$SPG->SetLatticeParameters($a,$b,$c,$alpha,$beta,$gamma);
	$SPG->SetLatticeSystem($lattice);
	$this->{"_symmetry_cell_setting"} = $SPG->LatticeSystem();

	$this->{"nTranslation"} = $SPG->AnalyzeTranslation();

#	my $nSymmetryOperation = 0;
	for(my $i = 1 ; ; $i++) {
		my $val = &GetValue($this, "_symmetry_equiv_pos_as_xyz", $i);
		$val = &GetValue($this, "_space_group_symop_operation_xyz", $i) if($val eq '');
		last if(!defined $val or $val eq '');
		$SPG->AddSymmetryOperation($val);
#		$nSymmetryOperation++;
	}
	$this->{"nSymmetryOperation"} = $SPG->nSymmetryOperation();
#print "a$LF";
	return;
}

sub AddSymmetryOperation
{
	my ($this,$symop) = @_;

	my $nSymmetryOperation = $this->{"nSymmetryOperation"};
	$nSymmetryOperation++;
#print "Add: i=$nSymmetryOperation: $symop<br>";
	$this->SetContent("_symmetry_equiv_pos_as_xyz", $nSymmetryOperation, $symop);
	return $this->{"nSymmetryOperation"} = $nSymmetryOperation;
}

sub GetCSpaceGroup
{
	my ($this) = @_;

	my $a = &GetValue($this, "_cell_length_a");
	my $b = &GetValue($this, "_cell_length_b");
	my $c = &GetValue($this, "_cell_length_c");
	my $alpha = &GetValue($this, "_cell_angle_alpha");
	my $beta  = &GetValue($this, "_cell_angle_beta");
	my $gamma = &GetValue($this, "_cell_angle_gamma");
	my $lattice = $this->{"_symmetry_cell_setting"};

	my $SPG = new SpaceGroup();
	my ($SPGName, $iSPG) = &GetSpaceGroup($this);
	$SPG->SetSPGName($SPGName);
	$SPG->SetiSPG($iSPG);
	$SPG->SetLatticeParameters($a,$b,$c,$alpha,$beta,$gamma);
	$SPG->SetLatticeSystem($lattice);
	$SPG->AnalyzeTranslation();
	for(my $i = 1 ; ; $i++) {
		my $val = &GetValue($this, "_symmetry_equiv_pos_as_xyz", $i);
		$val = &GetValue($this, "_space_group_symop_operation_xyz", $i) if($val eq '');
		last if(!defined $val or $val eq '');
		$SPG->AddSymmetryOperation($val);
	}
	return $SPG;
}

sub SetContent
{
	my ($this, $key, $idx, $content) = @_;

	if(@_ == 3) {
		$content = $idx;
		$idx = '';
	}
	$key = ${key} . "[$idx]" if($idx ne '' and $idx > 0);
	$this->{$key} = $content;
	return $content;
}

sub GetContent
{
	my ($this, $key, $idx, $defvalue) = @_;

	$key = ${key} . "[$idx]" if(defined $idx and $idx ne '' and $idx > 0);
	my $val = $this->{$key};
#print "key: $key\n";
#print "val: $val\n";
#print "keys: ", keys %$this, "\n";
#print "v1: $val => ";
	$val =~ s/^(\')(.*)(\1)$/$2/ if(defined $val);
#print "$val => ";
	$val = $defvalue if(defined $val and $val eq '');
#print "$val<br>\n";
	return $val;
}

sub GetValue
{
	my ($this, $key, $idx) = @_;

	my $val = &GetContent($this, $key, $idx);
#数値の最後の(誤差)を削除
	$val =~ s/^([+-\.\deE]+)\s*(\(.*?\))$/$1/ if(defined $val);
#両端の引用符を削除
	$val =~ s/^\s*(['"])(.*)\1\s*$/$2/ if(defined $val);
	
	return $val;
}

sub SetVolume
{
	my ($this,$v) = @_;
	return $this->SetContent("_cell_volume", $v);
}

sub Volume
{
	my ($this) = @_;
	return $this->GetValue("_cell_volume");
}

sub SetLatticeParameters
{
	my ($this,$a,$b,$c,$alpha,$beta,$gamma) = @_;
	$a = Utils::Round($a, 6);
	$b = Utils::Round($b, 6);
	$c = Utils::Round($c, 6);
	$alpha = Utils::Round($alpha, 6);
	$beta  = Utils::Round($beta , 6);
	$gamma = Utils::Round($gamma, 6);

	$this->SetContent("_cell_length_a", $a);
	$this->SetContent("_cell_length_b", $b);
	$this->SetContent("_cell_length_c", $c);
	$this->SetContent("_cell_angle_alpha", $alpha);
	$this->SetContent("_cell_angle_beta", $beta);
	$this->SetContent("_cell_angle_gamma", $gamma);

	return ($a,$b,$c,$alpha,$beta,$gamma);
}

sub LatticeParameters
{
	my ($this) = @_;

	my $a = &GetValue($this, "_cell_length_a");
	my $b = &GetValue($this, "_cell_length_b");
	my $c = &GetValue($this, "_cell_length_c");
	my $alpha = &GetValue($this, "_cell_angle_alpha");
	my $beta  = &GetValue($this, "_cell_angle_beta");
	my $gamma = &GetValue($this, "_cell_angle_gamma");

	return ($a,$b,$c,$alpha,$beta,$gamma);
}

sub GetSpaceGroup
{
	my ($this) = @_;

	my $SPGName = &GetValue($this, "_symmetry_space_group_name_H-M");
	$SPGName = &GetValue($this, "_space_group_name_H-M_alt") if($SPGName eq '');
	my $iSPG = &GetValue($this, "_symmetry_Int_Tables_number");
	$iSPG = &GetValue($this, "_space_group_IT_number") if($iSPG eq '');
	if($iSPG eq '') {
		$iSPG = 1;
		$SPGName = 'P 1';
	}
	return ($SPGName, $iSPG);
}

sub SetSpaceGroup
{
	my ($this, $SPGName, $iSPG) = @_;

	$this->SetContent("_symmetry_space_group_name_H-M", $SPGName);
	$this->SetContent("_symmetry_Int_Tables_number", $iSPG);

#	$this->{"nTranslation"} = $this->AnalyzeTranslation();

	return ($SPGName, $iSPG);
}

sub SetSpaceGroupInformation
{
	my ($this, $SPG) = @_;

	$this->SetSpaceGroup($SPG->SPGName(), $SPG->iSPG());
	my $nSymmetryOperation = $SPG->nSymmetryOperation();
	for(my $i = 0 ; $i < $nSymmetryOperation ; $i++) {
		my $symop = $SPG->SymmetryOperation($i);
		$this->AddSymmetryOperation($symop);
	}
}

sub LatticeSystem
{
	my ($this) = @_;

	return &GetValue($this, "_symmetry_cell_setting");
}

sub nTranslation
{
	my ($this) = @_;

	return &GetValue($this, "nTranslation");
}

sub nSymmetryOperation
{
	my ($this) = @_;

	return &GetValue($this, "nSymmetryOperation");
}

sub SymmetryOperation
{
	my ($this, $i) = @_;
	my $val = &GetValue($this, "_symmetry_equiv_pos_as_xyz", $i);
#print "Get: i=$i: $val<br>";
	return $val;
}

sub nAtomType
{
	my ($this, $i) = @_;

	return &GetValue($this, "nAtomType");
}

sub SetnAtomType
{
	my ($this, $n) = @_;
	return $this->SetContent("nAtomType", $n);
}

sub GetCAtomType
{
	my ($this, $i) = @_;

	my ($atomname, $charge) = $this->AtomType($i);
	my $atom = new AtomType();
	$atom->SetAtomName($atomname);
	$atom->SetCharge($charge); 
	return $atom;
}

sub GetCAtomTypeList
{
	my ($this) = @_;
	my $natom = $this->nAtomType();
	my @atomlist;
	for(my $i = 0 ; $i < $natom ; $i++)
	{
		my $atom = $this->GetCAtomType($i+1);
		push(@atomlist, $atom);
	}
	return @atomlist;
}

sub AtomType
{
	my ($this, $i) = @_;

	my $name   = &GetValue($this, "_atom_type_symbol", $i);
	my $charge = &GetValue($this, "_atom_type_oxidation_number", $i);
	if(!defined $name or $name eq '') {
		$name   = &GetValue($this, "_atom_site_label", $i);
		$name =~ s/[\d+-]+//g;
		$charge = 0.0;
	}
#print "name: $name, $charge<br>\n";
	return ($name, $charge);
}

sub AddAtomType
{
	my ($this, $atomtype) = @_;
	$atomtype = ucfirst $atomtype;
#print "AtomType: $atomtype\n";

	my $nAtomType = $this->nAtomType();
#print "na=$nAtomType\n";
	for(my $i = 0 ; ; $i++) {
		my $val = $this->GetContent("_atom_type_symbol", $i+1);
#		if(!defined $val or $val eq '') {
#			$val = &GetValue($this, "_atom_site_label", $i);
#			$val =~ s/[\d+-]+//g;
#		}
#print "i=$i: $val:$atomtype\n";
		last if($val eq '');
		return $nAtomType if($val eq $atomtype);
	}

	my ($name,$charge,$sign) = ($atomtype =~ /^(\D+)(\d+)?([\+-])?/);
	$charge = -$charge if($sign =~ /-/);
#print "name: $name - $charge - $sign<br>";
	$nAtomType++;
	$this->SetContent("_atom_type_symbol",           $nAtomType, $atomtype);
	$this->SetContent("_atom_type_oxidation_number", $nAtomType, $charge);
	$this->SetnAtomType($nAtomType);
	return $nAtomType;
}

#配列の引数は1からはじまる
sub SetAsymmetricAtomSiteVelocity
{
	my ($this, $i, $vx, $vy, $vz) = @_;
	$this->SetContent("Atom_site_velocity_x", $i, $vx);
	$this->SetContent("Atom_site_velocity_y", $i, $vy);
	$this->SetContent("Atom_site_velocity_z", $i, $vz);
	my $nSites = $this->nAsymmetricAtomSite();
	return $nSites;
}

sub AsymmetricAtomSiteVelocity
{
	my ($this, $i) = @_;
	my $vx = $this->GetContent("Atom_site_velocity_x", $i);
	my $vy = $this->GetContent("Atom_site_velocity_y", $i);
	my $vz = $this->GetContent("Atom_site_velocity_z", $i);
#print "CIF:AsymmetricAtomSiteVelocity: $i: ($vx, $vy, $vz)<br>\n";
	return ($vx, $vy, $vz);
}

sub AddAsymmetricAtomSiteWithVelocity
{
	my ($this, $label, $atomtype, $x, $y, $z, $occ,
		$vx, $vy, $vz) = @_;
	$vx = 0.0 if($vx eq '');
	$vy = 0.0 if($vy eq '');
	$vz = 0.0 if($vz eq '');
	my $i = $this->AddAsymmetricAtomSite($label,$atomtype,$x,$y,$z,$occ);
	$this->SetAsymmetricAtomSiteVelocity($i, $vx, $vy, $vz);
#print "CIF: $i: ($x,$y,$z) ($vx,$vy,$vz)<br>\n";
}

#配列の引数は1からはじまる
sub AddAsymmetricAtomSite
{
	my ($this, $label, $atomtype, $x, $y, $z, $occ) = @_;
	my $nSites = $this->nAsymmetricAtomSite();
	$nSites++;
	$this->SetContent("_atom_site_label",       $nSites, $label);
	$this->SetContent("_atom_site_type_symbol", $nSites, $atomtype);
#print "t: $atomtype\n";
	$this->AddAtomType($atomtype);
	$this->SetContent("_atom_site_fract_x",     $nSites, $x);
	$this->SetContent("_atom_site_fract_y",     $nSites, $y);
	$this->SetContent("_atom_site_fract_z",     $nSites, $z);
	$this->SetContent("_atom_site_occupancy",   $nSites, $occ);
	$this->SetnAsymmetricAtomSite($nSites);
#print "n: $nSites ($label)<br>";
	return $nSites;
}


sub SetnAsymmetricAtomSite
{
	my ($this,$n) = @_;

	return $this->SetContent("nAsymmetricAtomSite", $n);
}

sub nAsymmetricAtomSite
{
	my ($this) = @_;

	return &GetValue($this, "nAsymmetricAtomSite");
}

sub GetCAsymmetricAtomSite
{
	my ($this, $i) = @_;

	my ($label,$type,$x,$y,$z,$occupancy) 
		= $this->AsymmetricAtomSite($i);
	my ($vx,$vy,$vz)
		= $this->AsymmetricAtomSiteVelocity($i);

	my $atom = new AtomSite();
	$atom->SetLabel($label);
	$atom->SetAtomName($type); 
	$atom->SetPosition($x, $y, $z); 
	$atom->SetOccupancy($occupancy); 
#print "v2: $i: ($vx,$vy,$vz)<br>\n";
	$atom->SetVelocity($vx, $vy, $vz);

#	my ($vx,$vy,$vz) = $atom->Velocity();
#print "v3: $i: ($vx,$vy,$vz)<br>\n";
	
	return $atom;
}

sub GetCAsymmetricAtomSiteList
{
	my ($this) = @_;
	my $natom = $this->nAsymmetricAtomSite();
	my @atomlist;
	for(my $i = 0 ; $i < $natom ; $i++)
	{
		my $atom = $this->GetCAsymmetricAtomSite($i+1);
		push(@atomlist, $atom);
	}
	return @atomlist;
}

sub AsymmetricAtomSite
{
	my ($this, $i) = @_;

	my $label = &GetValue($this, "_atom_site_label", $i);
	my $type  = &GetValue($this, "_atom_site_type_symbol", $i);
	if(!defined $type or $type eq '') {
		$type = &GetValue($this, "_atom_site_label", $i);
		$type =~ s/[\d+-]+//g;
	}
	my $x = &GetValue($this, "_atom_site_fract_x", $i);
	my $y = &GetValue($this, "_atom_site_fract_y", $i);
	my $z = &GetValue($this, "_atom_site_fract_z", $i);
	my $occupancy = &GetValue($this, "_atom_site_occupancy", $i);
	return ($label,$type,$x,$y,$z,$occupancy);
}

"""