import os
import sys
import openpyxl

from tklib.tkobject import tkObject
from tklib.tkutils import mprint, IsDir, IsFile, SplitFilePath, modify_path, delete_file
from tklib.tkutils import terminate, safe_getelement, pint, pfloat, getarg, getintarg, getfloatarg, safe_getelement
from tklib.tksci.tkmatrix import make_matrix1, make_matrix2
import tklib.tkre
from tklib.tkdatafile import tkDataFile

#=======================================
# Handling MS-Excel .xlsx files
#=======================================


class tkExcel(tkDataFile):
    def __init__(self, path = None, mode = 'r', OpenFile = 1, CloseFile = 0, data_only = True, **args):
#        super(tkFile, self).__init__(path, mode, **args)
        self.fp      = None
        self.path    = None
        self.mode    = None
        self.wb      = None
        self.ws      = None
        self.cur_row = 0
        self.cur_col = 0
        self.labelarray    = []
        self.datalistarray = []
        self.update(**args)

        self.path    = self.IfYes(path is not None, path, self.path)
        self.mode    = self.IfYes(mode is not None, mode, self.mode)

        self.labelarray    = []
        self.datalistarray = []

        if OpenFile and self.path is not None:
            self.open(self.path, self.mode, data_only)

        if CloseFile:
            self.close()

    def __del__(self):
#        self.close()
        pass

    def __str__(self):
        return self.ClassPath()

    def initialize(self, **args):
        if self.fp:
            self.close()

        if self.wb:
            self.close()
            self.wb = None

        self.fp      = None
        self.path    = None
        self.mode    = None
        self.wb      = None
        self.ws      = None
        self.cur_row = 0
        self.cur_col = 0
        self.labelarray    = []
        self.datalistarray = []

        self.update(**args)
        
    def open(self, path = None, mode = None, data_only = True):
        self.initialize()
        self.SetPath(path, mode)

        if self.mode == 'w':
            if IsFile(path) and not delete_file(path):
                self.errormsg(sys._getframe().f_code.co_name, "Can not delete [{}]".format(self.path))
            self.wb = openpyxl.Workbook()
        elif self.mode == 'a':
#  UserWarning: Data Validation extension is not supported and will be removed がでるのでWarning抑制
            openpyxl.reader.excel.warnings.simplefilter('ignore')
            self.wb = openpyxl.load_workbook(self.path, data_only = data_only)
        else:
#  UserWarning: Data Validation extension is not supported and will be removed がでるのでWarning抑制
            openpyxl.reader.excel.warnings.simplefilter('ignore')
            self.wb = openpyxl.load_workbook(self.path, data_only = data_only)

        self.ws = self.wb.active
#        print("s=", self.path, self.mode, self.wb, self.ws)

        return self.wb

    def Open(self, path = None, mode = None, data_only = True):
        return self.open(path, mode, data_only)

    def close(self):
        if self.wb is not None:
            if self.mode == 'w' or self.mode == 'a':
                self.wb.save(self.path)
#            self.wb = None

    def Close(self):
        self.close()

    def set_path(self, path = None, mode = None):
        self.path = self.IfYes(path is not None, path, self.path)
        self.mode = self.IfYes(mode is not None, mode, self.mode)

    def SetPath(self, path = None, mode = None):
        return self.set_path(path = path, mode = mode)

    def max_row(self):
        return self.ws.max_row

    def max_column(self):
        return self.ws.max_column

    def get_sheet_by_name(self, title):
        return self.wb.get_sheet_by_name(title)

    def create_sheet(self, index = 0, title = None):
        self.wb.create_sheet(index = index, title = title)

    def remove_sheet(self, title):
        self.wb.remove_sheet(self.get_sheet_by_name(title))

    def set_sheet_name(self, title, sheet = None):
        if sheet is None:
            sheet = self.ws
        sheet.title = title

    def get_sheet(self, index):
        if type(index) is int:
            name = self.wb.get_sheet_names()[index]
        else:
            name = index
        return self.wb.get_sheet_by_name(name)                

    def get_column_letter(self, i):
        return openpyxl.utils.get_column_letter(i)

    def get_column_width(self, icolumn, sheet = None):
        if sheet is None:
            sheet = self.ws

        if type(icolumn) == int:
            icolumn = openpyxl.utils.get_column_letter(icolumn)
        return sheet.column_dimensions[icolumn].width

    def set_column_width(self, icolumn, width = None, sheet = None):
        if sheet is None:
            sheet = self.ws

        if type(icolumn) == int:
            icolumn = openpyxl.utils.get_column_letter(icolumn)
        sheet.column_dimensions[icolumn].width = width

    def get_column_hidden(self, icolumn, sheet = None):
        if sheet is None:
            sheet = self.ws

        if type(icolumn) == int:
            icolumn = openpyxl.utils.get_column_letter(icolumn)
        return sheet.column_dimensions[icolumn].hidden

    def set_column_hidden(self, icolumn, hidden = None, sheet = None):
        if sheet is None:
            sheet = self.ws

        if type(icolumn) == int:
            icolumn = openpyxl.utils.get_column_letter(icolumn)
        sheet.column_dimensions[icolumn].hidden = hidden

    def set_cell_font(self, row, column, sheet = None, font_color = '000000', 
                        italic = False, bold = False, strike = False, underline = None,
                        size = None, font_name = None,
                        vertAlign = 'baseline'):
        if sheet is None:
            sheet = self.ws

        font = openpyxl.styles.fonts.Font(
                    color = font_color, bold = bold, italic = italic, 
                    strike = strike, underline = underline,
                    size = size, name = font_name,
                    vertAlign = vertAlign)
        cell = sheet.cell(row = row, column = column)
        cell.font = font

    def set_cell_fill(self, row, column, sheet = None, fgcolor = 'FFFFFF', bgcolor = 'FFFFFF', pattern = 'solid'):
        if sheet is None:
            sheet = self.ws

        fill = openpyxl.styles.PatternFill(patternType = pattern, fgColor = fgcolor, bgColor = bgcolor)
        cell = sheet.cell(row = row, column = column)
        cell.fill = fill

    def get_irow_from_label(self, label, column_org = 1, row_org = 1, sheet = None):
        if sheet is None:
            sheet = self.ws

        for i in range(row_org, sheet.max_row + 1):
            s = sheet.cell(row = i, column = column_org).value
            if type(s) is int or type(label) is int:
                if pint(s) == pint(label):
                    return i
            elif type(s) is float or type(label) is float:
                if pfloat(s) == pfloat(label):
                    return i
            else:
                if s == label:
                    return i

        return None

    def get_icolumn_from_label1(self, label, column_org = 1, row_org = 1, sheet = None):
        if sheet is None:
            sheet = self.ws

        for i in range(column_org, sheet.max_column + 1):
            s = sheet.cell(row = row_org, column = i).value
            if type(s) is int or type(label) is int:
                if pint(s) == pint(label):
                    return i
            elif type(s) is float or type(label) is float:
                if pfloat(s) == pfloat(label):
                    return i
            else:
                if s == label:
                    return i

        return None

    def get_icolumn_from_label(self, label, column_org = 1, row_org = 1, sheet = None):
        if type(label) is list or type(label) is tuple:
            ilist = []
            for s in label:
                val = self.get_icolumn_from_label(s, column_org = column_org, row_org = row_org, sheet = sheet)
                ilist.append(val)
            return ilist
        else:
            return self.get_icolumn_from_label1(label, column_org = column_org, row_org = row_org, sheet = sheet)
        
    def find_val_in_row(self, val, irow, column_org = 1, sheet = None):
        return self.get_icolumn_from_label(val, column_org = column_org, row_org = irow, sheet = sheet)
        
    def find_val_in_column(self, val, icolumn, row_org = 1, sheet = None, auto_add = False):
        if sheet is None:
            sheet = self.ws

        ilast_vald_data = None
        for irow in range(row_org, sheet.max_row + 1):
            s = sheet.cell(row = irow, column = icolumn).value
            if s == val:
                return irow
            if s is not None and s != '':
                ilast_vald_data = irow

        if auto_add:
            ilast_vald_data += 1
            sheet.cell(row = ilast_vald_data, column = icolumn).value = val
            return ilast_vald_data

        else:
            return None

# cell = 'A1' to invalidate the freeze
    def freeze_panes(self, cell = 'A1', sheet = None):
        if sheet is None:
            sheet = self.ws
        sheet.freeze_panes = cell

    def store_column_widths(self, sheet = None, column_org = 1, row_org = 1):
        if sheet is None:
            sheet = self.ws

        max_column = sheet.max_column
        self.width_list = []
        self.width_dict = {}
        for icol in range(column_org, sheet.max_column+1):
            w = self.get_column_width(icol, sheet = sheet)
            label = sheet.cell(row = row_org, column = icol).value
            self.width_list.append(w)
            self.width_dict[label] = w

        return self.width_list, self.width_dict

    def restore_column_widths(self, sheet = None, column_org = 1, row_org = 1, by_label = True):
        if sheet is None:
            sheet = self.ws
            return None

        if by_label:
            for label in self.width_dict.keys():
                icol = self.get_icolumn_from_label(label, sheet = sheet, column_org = column_org, row_org = row_org)
                self.set_column_width(icol, width = self.width_dict[label], sheet = sheet)
        else:
            max_column = sheet.max_column
            for icol in range(column_org, sheet.max_column+1):
                self.set_column_width(icol, width = self.width_list[icol], sheet = sheet)

    def store_column_hidden(self, sheet = None, column_org = 1, row_org = 1):
        if sheet is None:
            sheet = self.ws

        max_column = sheet.max_column
        self.hidden_list = []
        self.hidden_dict = {}
        for icol in range(column_org, sheet.max_column+1):
            h = self.get_column_hidden(icol, sheet = sheet)
            label = sheet.cell(row = row_org, column = icol).value
            self.hidden_list.append(h)
            self.hidden_dict[label] = h

        return self.hidden_list, self.hidden_dict

    def restore_column_hidden(self, sheet = None, column_org = 1, row_org = 1, by_label = True):
        if sheet is None:
            sheet = self.ws
            return None

        if by_label:
            for label in self.hidden_dict.keys():
                icol = self.get_icolumn_from_label(label, sheet = sheet, column_org = column_org, row_org = row_org)
                self.set_column_hidden(icol, hidden = self.hidden_dict[label], sheet = sheet)
        else:
            max_column = sheet.max_column
            for icol in range(column_org, sheet.max_column+1):
                self.set_column_hidden(icol, hidden = self.hidden_list[icol], sheet = sheet)

    def insert_rows(self, irow):
        self.ws.insert_rows(irow)

    def insert_cols(self, icol, sheet = None, label = None, check_exist = False, 
                        column_org = 1, row_org = 1,
                        font = None, alignment = None, fill = None, border = None,
                        width = None, restore_width = True, restore_hidden = True):
        if sheet is None:
            sheet = self.ws

        if restore_width:
            self.store_column_widths(sheet = sheet, column_org = column_org, row_org = row_org)
        if restore_hidden:
            self.store_column_hidden(sheet = sheet, column_org = column_org, row_org = row_org)

        if check_exist and label is not None:
            idx = self.get_icolumn_from_label(label, sheet = sheet, column_org = column_org, row_org = row_org)
            if idx is None:
                sheet.insert_cols(icol)
                sheet.cell(row = row_org, column = icol).value = label

        idx = self.get_icolumn_from_label(label, sheet = sheet, column_org = column_org, row_org = row_org)

        if width is not None:
            self.set_column_width(icol, width = width, sheet = sheet)

        max_row = sheet.max_row

        if font is not None:
            if type(font) is int:
                if font >= icol:
                    isrc = font + 1
                else:
                    isrc = font
                
                for irow in range(row_org, max_row+1):
                    fo = sheet.cell(row = irow, column = isrc).font.copy()
                    sheet.cell(row = irow, column = idx).font = fo
            else:
                for irow in range(row_org, max_row+1):
                    sheet.cell(row = irow, column = idx).font = font

        if alignment is not None:
            if type(alignment) is int:
                if alignment >= icol:
                    isrc = alignment + 1
                else:
                    isrc = alignment
                for irow in range(row_org, max_row+1):
                    al = sheet.cell(row = irow, column = isrc).alignment.copy()
                    sheet.cell(row = irow, column = idx).alignment = al
            else:
                for irow in range(row_org, max_row+1):
                    sheet.cell(row = irow, column = idx).alignment = alignment

        if fill is not None:
            if type(fill) is int:
                if fill >= icol:
                    isrc = fill + 1
                else:
                    isrc = fill
                for irow in range(row_org, max_row+1):
                    fi = sheet.cell(row = irow, column = isrc).fill.copy()
                    sheet.cell(row = irow, column = idx).fill = fi
            elif type(fill) is str:
                for irow in range(row_org, max_row+1):
                    self.set_cell_fill(irow, idx, sheet = sheet, fgcolor = fill, bgcolor = fill, pattern = 'solid')
            else:
                for irow in range(row_org, max_row+1):
                    sheet.cell(row = irow, column = idx).fill = fill

        if border is not None:
            if type(border) is int:
                if border >= icol:
                    isrc = border + 1
                else:
                    isrc = border
                for irow in range(row_org, max_row+1):
                    bo = sheet.cell(row = irow, column = isrc).border.copy()
                    sheet.cell(row = irow, column = idx).border = bo
            else:
                for irow in range(row_org, max_row+1):
                    sheet.cell(row = irow, column = idx).border = border

        if restore_width:
            self.restore_column_widths(sheet = sheet, 
                    column_org = column_org, row_org = row_org, by_label = True)
        if restore_hidden:
            self.restore_column_hidden(sheet = sheet, 
                    column_org = column_org, row_org = row_org, by_label = True)

        return icol

    def delete_rows(self, irow):
        self.ws.delete_rows(irow)

    def delete_cols(self, icol):
        self.ws.delete_cols(icol)

    def get_row(self, irow):
        self.cur_row = irow + 1
        self.cur_col = 0
        return self.ws[irow+1]
        
    def get_val(self, irow, icol):
        return self.ws.cell(irow+1, icol+1).value

    def Get(self, irow, icol):
        return self.get_val(irow, icol)

    def append(self, datas):
        self.ws.append(datas)
        self.cur_row = self.max_row() + 1
        self.cur_col = 0

    def Append(self, datas):
        self.append(datas)

    def save(self, path, workbook = None):
        self.SetPath(path, mode = 'w')
        if workbook is None:
            workbook = self.wb

        return workbook.save(path)

    def write(self, irow, icol, val):
        if type(val) is dict:
            self.ws.cell(irow + 1, icol + 1).value = val["val"]
            if val.get("number_format", None) is not None:
                self.ws.cell(irow + 1, icol + 1).number_format = val["number_format"]
            if val.get("horizontal_allignment", None) is not None:
                if val["horizontal_allignment"] == '<':
                    self.ws.cell(irow + 1, icol + 1).alignment = Alignment(horizontal = "leftContinuous")
                elif val["horizontal_allignment"] == '^':
                    self.ws.cell(irow + 1, icol + 1).alignment = Alignment(horizontal = "centerContinuous")
                elif val["horizontal_allignment"] == '>':
                    self.ws.cell(irow + 1, icol + 1).alignment = Alignment(horizontal = "rightContinuous")
        else:
            self.ws.cell(irow + 1, icol + 1).value = val

    def Write(self, irow, icol, val):
        self.write(irow, icol, val)

    def print(self, *vals, end = '\n'):
        if type(vals[0]) is list or type(vals[0]) is tuple:
            for v in vals[0]:
                self.Write(self.cur_row, self.cur_col, v)
                self.cur_col += 1
        else:
            for i in range(len(vals)):
                self.Write(self.cur_row, self.cur_col, vals[i])
                self.cur_col += 1

        if end == '\n':
            self.cur_row += 1
            self.cur_col = 0
#            print("r=", self.cur_row, self.cur_col)
        else:
            pass
#        print("r=", self.cur_row, self.cur_col)

    def Print(self, *vals, end = '\n'):
        self.print(*vals, end)
        
    def load_workbook(self, path = None):
        self.SetPath(path = path, mode = 'w')

#  UserWarning: Data Validation extension is not supported and will be removed がでるのでWarning抑制
        openpyxl.reader.excel.warnings.simplefilter('ignore')
        book = openpyxl.load_workbook(self.path)
        self.wb = book

        return book

    def sheetnames(self, path = None):
#  UserWarning: Data Validation extension is not supported and will be removed がでるのでWarning抑制
        openpyxl.reader.excel.warnings.simplefilter('ignore')
        book = self.load_workbook(path)
        return book.sheetnames

    def read_sheet(self, path = None, sheet_name = None, mode = 'r', data_only = True):
        self.SetPath(path = path, mode = mode)

        if self.wb is None:
            wb = self.Open(self.path, mode = mode, data_only = data_only)
            was_opened = False
        else:
            was_opened = True
        
        if sheet_name is not None:
            try:
                idx = int(sheet_name)
                ws = self.wb.worksheets[idx]
            except:
                ws = self.wb[sheet_name]
        else:
            ws = self.wb.active

        cells = []
        irow = 0
        for row in ws:
            cells.append([])
            icol = 0
            for cell in row:
                cells[irow].append(cell)
                icol += 1
            irow += 1

        if was_opened:
            self.Close()

        self.ws = ws

        return cells

    def Read_sheet(self, path = None, sheet_name = None, mode = 'r', data_only = True):
        return self.read_sheet(path = path, sheet_name = sheet_name, mode = mode, data_only = data_only)

    def get_list_from_cells(self, cells = None, debug = False):
        header = []
        for cell in cells[0]:
            header.append(cell.value)
#        print("header=", header)

        nrow = len(cells)
        ncol = len(cells[0])

        if debug:
            print("Read_sheet_list: len(cells)=", len(cells))
            print("Read_sheet_list: len(cells[0])=", len(cells[0]))

        datalist = make_matrix2(ncol, nrow - 1, defval = None)
        for irow in range(nrow-1):
            row  = self.get_row(irow+1)
#            vals = []
            for icol in range(ncol):
#                val = cells[irow+1][icol].value
                val = self.get_val(irow+1, icol)
                datalist[icol][irow] = val
#            datalist.append(vals)

        if debug:
            print("Read_sheet_list: len(datalist)=", len(datalist))
            print("Read_sheet_list: len(datalist[0])=", len(datalist[0]))

#        print("datalist=", datalist)
#        for i in range(nrow):
#            for j in range(ncol):
#                print(datalist[i][j], " ", end = '')
#            print("")

        return header, datalist

    def read_sheet_list(self, path= None, sheet_name = None, mode = 'r', data_only = True, debug = False):
        cells = self.read_sheet(path = path, sheet_name = sheet_name, mode = mode, data_only = data_only)

        header, datalist = self.get_list_from_cells(cells = cells, debug = debug)

        return header, datalist

    def Read_sheet_list(self, path = None, sheet_name = None, mode = 'r', data_only = True, debug = False):
        self.read_sheet_list(path = path, sheet_name = sheet_name, mode = mode, data_only = data_only, debug = debug)

    def get_dict_from_datalist(self, datalist = None, keys = None, allow_overlapped_keys  = True):
        nrow = len(datalist[0])
        ncol = len(keys)

        dict = {}
        for i in range(ncol):
            field_name = keys[i]
            dict[field_name] = []

        red_field = {}
        for icol in range(ncol):
            field_name = keys[icol]
            if not allow_overlapped_keys and red_field.get(field_name, None) is not None:
                continue

            red_field[field_name] = 1

            for irow in range(nrow):
                dict[field_name].append(datalist[icol][irow])

        return dict

    def read_sheet_dict(self, path = None, sheet_name = None, allow_overlapped_keys = True, debug = False):
        keys, datalist = self.read_sheet_list(path = path, sheet_name = sheet_name, debug = debug)
#        sheetnames = self.sheetnames()
        sheetnames = self.wb.sheetnames

        if debug:
            print("Read_sheet_dict: len(datalist)=", len(datalist))
            print("Read_sheet_dict: len(datalist[0])=", len(datalist[0]))
            exit()
        
        dict = self.get_dict_from_datalist(datalist = datalist, keys = keys, allow_overlapped_keys = allow_overlapped_keys)

        return dict, sheetnames, keys, datalist

    def Read_sheet_dict(self, path = None, sheet_name = None, allow_overlapped_keys = True, debug = False):
        return self.read_sheet_dict(path = path, sheet_name = sheet_name, 
                        allow_overlapped_keys = allow_overlapped_keys, debug = debug)

    def get_matrix_from_cells(self, cells = None):
        header = []
        for cell in cells[0]:
            header.append(cell.value)
#        print("header=", header)
        matrix = []
        for irow in range(1, len(cells)):
            row = self.get_row(irow)
            vals = []
            for cell in row:
                vals.append(cell.value)
            matrix.append(vals)
#        print("datalist=", datalist)

    def read_sheet_matrix(self, sheet_name = None):
        cells = self.read_sheet()
        header, matrix = self.get_matrix_from_cells(cells = cells)

        return header, matrix

    def Read_sheet_matrix(self, sheet_name = None):
        return self.read_sheet_matrix(sheet_name = sheet_name)

    def read_minimum_matrix(self, close_fp = False):
        labels   = []
        row = self.get_row(0)
        for cell in row:
            val = cell.value
            if val is None:
                break

            if type(val) is str:
                val.strip()
            labels.append(val)

        ncol = len(labels)
        datalist = make_matrix1(ncol, type = 'list')
        irow = 1
        while 1:
            row = self.get_row(irow)
            if row is None or row[0].value is None:
                break

            for icol in range(ncol):
                try:
                    val = pfloat(row[icol].value)
                except:
                    val = None
                datalist[icol].append(val)

            irow += 1

        if close_fp:
            self.Close()

        self.labelarray    = labels
        self.datalistarray = datalist

        return labels, datalist

    def Read_minimum_matrix(self, close_fp = False):
        return self.read_minimum_matrix(close_fp = close_fp)

def main():
    path = "a.xlsx"
    xls = tkExcel(path, 'w')
    xls.Write(0, 0, 'a00')
    xls.Write(1, 1, 'a11')
    xls.Close()

    xls = tkExcel(path, 'a')
    xls.Write(0, 1, 'a01')
    xls.Write(1, 1, 'a22')
    xls.Close()

    xls = tkExcel(path, 'r')
    print("00=", xls.Get(0, 0))
    print("01=", xls.Get(0, 1))
    print("10=", xls.Get(1, 0))
    print("11=", xls.Get(1, 1))
    xls.Close()

    path = "b.xlsx"
    xls = tkExcel(path, 'w')
    xls.print([0, 1, 2, 'aa'])
    xls.print([3, 0, 1, 2, 'bb'])
    xls.print([3, 'a&'], end = '')
    xls.print([3, 5])
    xls.Close()

    
if __name__ == "__main__":
    main()
