import os
import os.path
import platform
import sys
import shutil
import datetime
import glob
import re
import csv
import unicodedata
import numpy as np
from matplotlib.markers import TICKLEFT, TICKRIGHT, TICKUP, TICKDOWN, CARETLEFT, CARETRIGHT, CARETUP, CARETDOWN

from tklib.tkobject import tkObject, _analyze_varstr
import tklib.tkre as tkre


colors = ['black', 'red', 'blue', 'darkgreen', 'darkorange', 'navy', 
          'slategray', 'hotpink', 'olive', 'chocolate', 'magenta', 
          'green', 'yellow', 'cyan']

markers = [
    ('.', 'point'),
    (',', 'pixel'),
    ('o', 'circle'),
    ('v', 'triangle_down'),
    ('^', 'triangle_up'),
    ('<', 'triangle_left'),
    ('>', 'triangle_right'),
    ('1', 'tri_down'),
    ('2', 'tri_up'),
    ('3', 'tri_left'),
    ('4', 'tri_right'),
    ('8', 'octagon'),
    ('s', 'square'),
    ('p', 'pentagon'),
    ('*', 'star'),
    ('h', 'hexagon1'),
    ('H', 'hexagon2'),
    ('+', 'plus'),
    ('x', 'x'),
    ('D', 'diamond'),
    ('d', 'thin_diamond'),
    ('|', 'vline'),
    ('_', 'hline'),
    (TICKLEFT, 'tickleft'),
    (TICKRIGHT, 'tickright'),
    (TICKUP, 'tickup'),
    (TICKDOWN, 'tickdown'),
    (CARETLEFT, 'caretleft'),
    (CARETRIGHT, 'caretright'),
    (CARETUP, 'caretup'),
    (CARETDOWN, 'caretdown'),
    ('None', 'nothing'),
    (None, 'nothing'),
    (' ', 'nothing'),
    ('', 'nothing')
    ]


def pdb():
    print("Enter to pdb: see https://docs.python.org/ja/3/library/pdb.html")
    print("   h: help")
    print("   p, pp: print var")
    print("   what is: print type of var")
    import pdb; pdb.set_trace()

def analyze_varstr(str):
    return _analyze_varstr(str)

def split_optstr(str):
    try:
        val, id = str.split(':')
        return pfloat(val), pint(id)
    except:
        return str, None

def quote_command_if_space(s, special_chars = ""):
    if s is None:
        return s
    
    if ' ' not in s and not re.search(rf"[{special_chars}]", s):
        pass
    elif len(s) >= 1 and '/-' in s:
        pass
    elif re.match(r"\s*[\"\'].*[\"\']", s):
        pass
    else:
        s = f'"{s}"'
    return s

def convert_by_vars(s, vars, prefix = r'\$', bra = r'\(', ket = r'\)'):
    idebug = 0

    if idebug >= 1:
        print("")
        print("In tkutils.convert_by_vars:")
        print(f"  vars={vars}")

    ret = ''
    rest = s
    if idebug >= 2:
        print(f"89 rest={rest}")
        print(f"90 ret={ret}")
    while 1:
        if idebug >= 2:
            print(f"93 rest={rest}")
            print(f"94 ret={ret}")

        if prefix != '':
# Separate by '$'
            m = re.match(rf'(.*?){prefix}(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
#            m = re.match(r'(.*?)\$(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
            if not m:
                if idebug >= 1:
                    print(f" not matched for $: rest={rest}")
                return ret + rest

            if idebug >= 1:
                print(f" match for $: rest={rest}")
                print(f"              ret={ret}")

            g = m.groups()
            pre = g[0]
            n   = len(pre)
            rest = g[1]
            if idebug >= 2:
                print(f"  pre={pre} rest={rest}")

# Case rest = 'abc\$def', should be converted to 'abc$def'
            if n > 0 and pre[n-1] == '\\':
                ret += pre[0:n-1] + '$'
            else:
                ret += pre

# Case rest = 'abc$i def', should be converted to 'abc[ini_path] def'
            m = re.match(r'([a-zA-Z])([^\w_].*)?$', rest, flags = re.MULTILINE | re.DOTALL)
            if m:
                if debug >= 1:
                    print(f"121: not matched")
                g = m.groups()
                n = len(g)
                varname = g[0].lower()
                rest = g[1]
                if rest is None:
                    rest = ''

                val = vars.get(varname, None)
#                    print(f"   1{var=} {val=}")
                if val is None:
                    app.show_error_dialog(f"Invalid varname [${varname}]")
                    val = f"??{varname}??"

                ret += val
                if idebug >= 2:
                    print(f"   1 ret={ret}")

                continue
        else:
            m = re.match(rf'(.*?)({bra}.*)$', rest, flags = re.MULTILINE | re.DOTALL)
            if not m:
                if idebug >= 1:
                    print(f" not matched for $: rest={rest}")
                return ret + rest

            g = m.groups()
            ret += g[0]
            rest = g[1]

# Case rest = 'abc$(tkProgRoot)def', should be converted to 'abc[tkProgRoot]def'
        m = re.match(rf'{bra}(\w+){ket}(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
#        m = re.match(r'\((\w+)\)(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
        if m:
            if idebug >= 1:
                print(f"146: not matched rest={rest}")
            g = m.groups()
            varname = g[0].lower()
            rest = g[1]

#            print(f"{varname=}")
            val = vars.get(varname, None)
            if idebug >= 2:
                 print(f"   3 varname={varname} val={val}")
                 print(f"   3 vars={vars} val={val}")
            if val is None:
                print("")
                print(f"Warning in tkutils.convert_by_vars(): Invalid varname [{prefix}{bra}{varname}{ket}]")
                val = f"??{varname}??"

            ret += f"{val}"
            if idebug >= 2:
                print(f"   4 ret={ret}")

            continue

#      print(f"193 ret={ret}")
    return ret


def del_quote(s):
    n = len(s)
    if n <= 1:
        return s

    if s[0] == '"' and s[n-1] == '"':
        return s[1:n-1]
    if s[0] == "'" and s[n-1] == "'":
        return s[1:n-1]
    return s

def split_quoted_args(s):
#    arg_list = s.split()

    rest = s.strip()
    if rest == '':
        return rest

    args = []
    while 1:
        if rest[0] == '"':
            m = re.match(r'(.*?)\"\s+(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
            if not m:
                args.append(del_quote(rest))
                break

            a1 = m.groups()[0][1:]
            rest = m.groups()[1]
            args.append(a1)
        elif rest[0] == "'":
            m = re.match(r'(.*?)\'\s+(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
            if not m:
                args.append(del_quote(rest))
                break

            a1 = m.groups()[0][1:]
            rest = m.groups()[1]
            args.append(a1)
        else:
            m = re.match(r'(.*?)\s+(.*)$', rest, flags = re.MULTILINE | re.DOTALL)
            if not m:
                args.append(del_quote(rest))
                break

            a1 = m.groups()[0]
            rest = m.groups()[1]
            args.append(a1)

    return args

def split_two(s):
    s = s.strip()
#    print(f"250: s={s}")

    if len(s) == 0:
        return '', ''

    if s[0] == '"':
        m = re.match(r'(.*?)\"\s+(.*)$', s, flags = re.MULTILINE | re.DOTALL)
        if m:
            g = m.groups()
            s1 = g[0][1:]
            s2 = g[1]
#            print(f"260 s1={s1} s2={s2}")
            return s1, s2
    elif s[0] == "'":
        m = re.match(r'(.*?)\'\s+(.*)$', s, flags = re.MULTILINE | re.DOTALL)
        if m:
            g = m.groups()
            s1 = g[0][1:]
            s2 = g[1]
#            print(f"268 s1={s1} s2={s2}")
            return s1, s2

    m = re.match(r'(\S+)\s+(.*)\s*$', s, flags = re.MULTILINE | re.DOTALL)
    if not m:
#        print(f"273 s1={s}")
        return s, ''

    g = m.groups()
#    print(f"276 s1={g[0]} s2={g[1]}")

    return g[0], g[1]

def split_command_line(s):
    cmd, rest = split_two(s)
    args = split_quoted_args(rest)
    
    return cmd, rest, args

def mprint(*args, fp = None, **kwargs):
    print(*args, **kwargs)
    if fp:
        print(*args, file = fp, **kwargs)
        fp.flush()

def get_path_list(varname = 'PATH'):
    return getenv(varname, '').split(os.pathsep)

def add_path(path, varname = 'PATH'):
    os_name = get_os()
    sep = os.pathsep
    paths = get_path_list(varname = varname)

    path_list = [path]
    is_found = False
    if os_name == 'Linux':
        for p in paths:
            if p == path:
                return os.environ.get(varname, "")
            path_list.append(p)
    else:
        path = path.lower()
        for p in paths:
            if p.lower() == path:
                return os.environ.get(varname, "")
            path_list.append(p)

#    print("path=", path)
#    print("path_list=", path_list)
    os.environ[varname] = sep.join(path_list)
    
    return os.environ.get(varname, "")

def normalize_str(str, conv_table = None, target_char = ' ', unicodedata_mode = 'NFKC', ):
    if conv_table is None:
        conv_table = str.maketrans({
              '\u3000': target_char
            , ' '     : target_char
            , '\t'    : target_char
            , '\r'    : target_char
            , ', '    : target_char
            , ','     : target_char
            , '，'    : target_char
            })

    str = str.translate(conv_table)

    str = unicodedata.normalize(unicodedata_mode, str)

    return str

def normalize_name(name, conv_table, unicodedata_mode = 'NFKC', upper_mode = 'upper'):
    name = normalize_str(name, conv_table = conv_table, unicodedata_mode = unicodedata_mode)
    name = name.replace('  ', ' ')
    name = name.strip()
    if upper_mode == 'upper':
        name = name.upper()
    elif upper_mode == 'lower':
        name = name.lower()
    elif upper_mode == 'capitalize':
        name = name.capitalize()
    elif upper_mode == 'title':
        name = name.title()

    return name

def add_dict(var, key1, key2 = None):
    if key2 is None:
        if var.get(key1, None) is None:
            var[key1] = 1
        else:
            var[key1] += 1
    else:
        if var.get(key1, None) is None:
            var[key1] = {}
        if var[key1].get(key2, None) is None:
            var[key1][key2] = 1
        else:
            var[key1][key2] += 1

def get_os_info(terse = False):
    return platform.system(), platform.platform(terse = terse),\
              platform.release(), platform.version()

# macOS: 'Darwin', Windows: 'Windows', Linux: 'Linux'
def get_os():
    return platform.system()

def is_windows():
    if platform.system() == 'Windows':
        return platform.version()
    return False

def is_mac():
    if platform.systmem() == 'Darwin':
        return platform.version()
    return False

def is_linux():
    if platform.systmem() == 'Linux':
        return platform.version()
    return False

def getenv(key, defval = None):
    return os.environ.get(key, defval)

def setenv(key, val = None):
    if val is None:
        if key in os.environ[key]:
            del os.environ[key]
            return
    os.environ[key] = val

def GetPathDelimiter():
    return os.sep

def get_path_separator():
    return os.sep

def get_PATH_env_separator():
    return os.pathsep

def get_ext_separator():
    return os.extsep

def is_dir(path):
    return os.path.isdir(path)

def IsDir(path):
    return os.path.isdir(path)

def is_file(path):
    return os.path.isfile(path)

def IsFile(path):
    return os.path.isfile(path)

def is_exist(path):
    return os.path.exists(path)

def Exists(path):
    return os.path.exists(path)

def make_path(dir, *args):
    return os.path.join(dir, *args)

def MakePath(dir, *args):
    return make_path(dir, *args)

def get_fullpath(path):
    return os.path.abspath(path)

def get_dirname(path):
    fullpath      = os.path.abspath(path)
    dirname       = os.path.dirname(fullpath)
    return dirname
    
def get_last_directory(path):
    dirname, basename, filebody, ext = split_file_path(path)
    last_dir = os.path.basename(os.path.normpath(dirname))
    return last_dir

def split_drive(path):
    dirname, basename, filebody, ext = split_file_path(path)
    last_dir = os.path.basename(os.path.normpath(dirname))
    drive = ''
    dirname_wo_drive = dirname
    m = re.match(r'([a-zA-Z]:)(.*)$', dirname)
    if m:
        drive = m.groups()[0]
        dirname_wo_drive = m.groups()[0]
    return drive, dirname_wo_drive

def split_file_path(path):
    if os.sep == '\\':
        path0 = tkre.Sub(r'\\', '/', path)
    else:
        path0 = path
    path0 = tkre.Sub('/$', '', path0)

#    dirname, basename, os.path.split(path)
    basename    = os.path.basename(path0)
    dirname     = os.path.dirname(path0)
    header, ext = os.path.splitext(path0)
    filebody    = os.path.basename(header)

    if os.path.isdir(path):
        ext = ''
    if os.sep == '\\':
        dirname = tkre.Sub('/', r'\\', dirname)

    return dirname, basename, filebody, ext

def SplitFilePath(path):
    return split_file_path(path)

def modify_path(path, addpath):
    dirname, basename, filebody, ext = split_file_path(path)
    return make_path(dirname, filebody + addpath)

def replace_path(path, template, ext_dict = {}):
    fullpath      = os.path.abspath(path)
    filename      = os.path.basename(fullpath)
    dirname       = os.path.dirname(fullpath)
    filebody, ext = os.path.splitext(filename)
    ext_dict.update({
        "fullpath": fullpath,
        "dirname": dirname,
        "filename": filename,
        "filebody": filebody,
        "ext": ext
        })

    return template.format(**ext_dict)

def get_file_masks(key = 'executables'):
    if is_windows():
        if key == 'executables':
            exts = getenv('PATHEXT', '').lower().split(';')
            return ["*{}".format(ext) for ext in exts] 
        return ["*.exe", "*.com", "*.bat", "*.vbs", "*.sh", "*.csh"]

    return ["*.*"]

def find_executable_path(filename, defval = None):
    paths = get_path_list()
    for path in paths:
#        print("path:", path)
        f = f"{path}{os.sep}{filename}"
#        print("  f:", f)
        if Exists(f):
#            print("    Exists")
            return f

        if is_windows():
            exts = getenv('PATHEXT', '').lower().split(';')
            for ext in exts:
                p = f'{f}{ext}'
#                print("  p:", p)
                if Exists(p):
#                    print("    Exists")
                    return p
    return defval

def search_executable_path(filenames, defval = None):
    for f in filenames:
        path = find_executable_path(f, defval = None)
        if path:
            return path
    return defval

def terminate(message = None, usage = None, pause = False):
    if message is not None:
        print("")
        print(message)

    if usage:
        print("")
        usage()

    print("")
    app.print("")
    if pause:
        input("Press ENTER to terminate >> ")
        app.print("")

    exit()

def sfmt(str, fmt):
    f = '{' + fmt + '}'
    return f.format(str).strip()

def safe_getelement(var, key, defval = None):
    try:
        return var[key]
    except:
        return defval

def pint(s, strict = True, defval = 0):
    try:
        return int(s)
    except:
        pass
    if s is None or strict:
        return defval

    list = tkre.Search(r'([\+\-]?[\d]+)', s)
    try:
        return int(list[1])
    except:
        return defval

def pfloat(s, strict = True, defval = 0.0):
    try:
        return float(s)
    except:
        pass
    if s is None or strict:
        return defval

    list = tkre.Search(r'([\+\-]?[\d\.eE]+)', s)
    try:
        return float(list[1])
    except:
        if defval is None:
            return s
        return defval

def pintfloat(s, strict = True, defval = 0.0):
    try:
        return int(s)
    except:
        pass

    return float(s, defval = defval, strict = strict)

def pconv(s, strict = True, defval = 0):
    if s is None:
        return defval

    try:
       return int(s)
    except:
        pass

    try:
       return float(s)
    except:
        return s

    return s

def pconv_by_type(s, type = 'str', defval = None, strict = False):
    if strict:
        if type == 'int':
            try:
                return int(s)
            except:
                return defval
        if type == 'float':
            try:
                return float(s)
            except:
                return defval
        return s
    else:
        if type == 'int':
            return pint(s, defval = defval, strict = strict)
        if type == 'float':
            return pfloat(s, defval = defval, strict = strict)
        return s

def getarg(position, defval = None):
    try:
        return sys.argv[position]
    except:
        return defval

def getfloatarg(position, defval = None):
    return pfloat(getarg(position, defval))

def getintarg(position, defval = None):
    return pint(getarg(position, defval))

def GetList(list, n, defval = 0.0):
    l = list.copy()
    for i in range(len(l), n):
        l.append(defval)
    return l[0:n]

def validate_error(y0, y1, eps, message):
    if y0 == 0.0:
        err = (y1 - y0) / y1
    else:
        err = (y1 - y0) / y0
    if abs(err) > eps:
        print("")
        print("{}Too large error {} vs {} error={:8.3g} > {:8.3g}".format(message, y0, y1, err, eps))
        print("")
        exit()

def read_csv_to_dict(path, delimiter = ',', newline = '', convert_float = 0, convert_int = 0):
    inf = []
    try:
        with open(path) as f:
            reader = csv.DictReader(f, delimiter = delimiter, newline = newline)
            for row in reader:
                inf.append(row)
        if convert_float:
            for i in range(len(inf)):
                for key, val in inf[i].items():
                     inf[i][key] = pfloat(inf[i][key], None)
        elif convert_int:
            for i in range(len(inf)):
                for key, val in inf[i].items():
                     inf[i][key] = pint(inf[i][key], None)
        return inf
    except:
        return 0

def read_csv(path, delimiter = ',', newline = '',
            first_header = 1, convert_float = 0, convert_int = 0):
    inf = []
    header = None
    try:
        with open(path) as f:
            reader = csv.reader(f, delimiter = delimiter, newline = newline)
            for row in reader:
                inf.append(row)
        if first_header:
            header = inf[0]
            inf    = inf[1:]
        if convert_float:
            for i in range(len(inf)):
                for j in range(len(inf[i])):
                     inf[i][j] = pfloat(inf[i][j], None)
        elif convert_int:
            for i in range(len(inf)):
                for j in range(len(inf[i])):
                     inf[i][j] = pint(inf[i][j], None)
        return header, inf
    except:
        return 0

def read_csv2(fname, delimiter = ','):
    print("")
    with open(fname) as f:
        fin = csv.reader(f, delimiter = delimiter)
        
        labels = next(fin)
        xlabel = labels[0]

# label行が 空文字 の場合、データとしては読み込まない
        ylabels = []
        for i in range(1, len(labels)):
            if labels[i] == '':
                break
            ylabels.append(labels[i])
        ny = len(ylabels)
        print("xlabel: ", xlabel)
        print("ylabels: ", ylabels)
        print("ny=", ny)

        x     = []
        ylist = []
        for i in range(ny):
            ylist.append([])
        
        for row in fin:
            x.append(pfloat(row[0]))
            for i in range(1, ny+1):
                v = pfloat(row[i])
                if v is not None:
                    ylist[i-1].append(v)
                else:
                    ylist[i-1].append(None)

    return [xlabel, *ylabels], [x, *ylist]

    
def save_csv(path, headerlist, datalist, is_print = 0):
    f = open(path, 'w')
    if not f:
        return 0

#    print("len=", len(datalist), len(datalist[0]))
    writer = csv.writer(f, delimiter=',', lineterminator='\n')
    writer.writerow(headerlist)
    for i in range(len(datalist[0])):
        dlist = [datalist[id][i] for id in range(len(datalist))]
        if is_print:
            try:
                dliststr = joinf(dlist, "%12.8g", ", ")
            except:
                dliststr = joinf(dlist, "%12s", ", ")
            print("  {:3d}: {}".format(i, dliststr))

        writer.writerow(dlist)

    f.close()
    return 1

def joinf(list, format, sep):
    s = format % (list[0])
    for i in range(1, len(list)):
        s += sep + format % (list[i])
    return s

def merge_attributes(tobj, sobj):
    if sobj is None:
        return tobj
    for key, val in sobj.__dict__.items():
        try:
            getattr(tobj, key)
        except:
            setattr(tobj, key, val)
    return tobj
    

def check_attributes(obj, isprint = 0, *args):
    list = []
    for key in args:
        try:
            getattr(obj, key)
        except:
            list.append(key)
    if len(list) == 0:
        return None
    else:
        if isprint:
            print("tkutils.check_attributes: The following attributes are not defined:",
                list)
        return list


def lvlprint(lprint, level, *args):
    if lprint >= level:
        print(*args)

def BuildCreationDateStr(date = None):
    if date == None:
        date = datetime.datetime.now()
    s = "%04d/%02d/%02d" % (date.year, date.month, date.day)
    return s

def copy_path(path, newpath):
    try:
        shutil.copy2(path, newpath, follow_symlinks=True)
        return 1
    except:
        return 0

def rename_file(path, newpath):
    try:
        os.rename(path, newpath)
        return 1
    except:
        return 0

def delete_file(path):
    try:
        os.remove(path)
        return 1
    except:
        return 0


