from scipy.optimize import minimize


#nelder-mead    Downhill simplex
#powell         Modified Powell
#cg             conjugate gradient (Polak-Ribiere method)
#bfgs           BFGS法
#newton-cg      Newton-CG
#trust-ncg      信頼領域 Newton-CG 法
#dogleg         信頼領域 dog-leg 法
#L-BFGS-B’ (see here)
#TNC’ (see here)
#COBYLA’ (see here)
#SLSQP’ (see here)
#trust-constr’(see here)
#dogleg’ (see here)
#trust-exact’ (see here)
#trust-krylov’ (see here)

method = "bfgs"


#==========================================
# Source parameters to be fitted
#==========================================
x0s = [0.0, 0.0]

# 数値微分する際の変数の微小変位
h = 0.01

maxiter = 100
tol = 1.0e-5


def minimize_func(x):
    return 2.0 * (x[0] - 3.0)**2 + (x[1] - 1.0)**4 + 2.0

# １次微分を定義するとcgやbfgsなどの勾配法を使える
def diff1(x):
    diff = x.copy()
    nvar = len(x)

    for i in range(nvar):
        xx = x.copy()
        xx[i] = x[i] - h
        ym = minimize_func(xx)
        
        xx[i] = x[i] + h
        yp = minimize_func(xx)
        diff[i] = (yp - ym) / 2.0 / h

    return diff


# callbackを使うと、最適化過程をモニターできる
# 引数は変数のリストだけが渡される
# 反復回数などはglobal変数で保持する
iter = 0
def callback(xk):
    global iter

    fmin = minimize_func(xk)
    print(f"callback {iter}: xk={xk}  func={fmin}")
    iter += 1


#==========================================
# Main routine
#==========================================
def main():
    global x0s, tol, maxiter
    
    print("")
    print(f"method {method}")
    print(f"initial values {x0s}")
    print(f"tol {tol}")

    print("")
    print("Minimize:")
    res = minimize(minimize_func, x0s, method = method, jac = diff1, callback = callback,
                tol = tol, options = {'maxiter':maxiter, "disp":True})
#    print("")
#    print(res)
    print("")
    if res.success:
        print(f"Function takes the minimum")
        print(f"   at y={res.fun}")
        print(f"   with x={res.x}")
        print(f" iteration: {res.nit}")
    else:
        print(f"Function did not converge")
        print(res)

if __name__ == "__main__":
    main()
